"""Calculate the cost of a Grok API call."""

from ..base.types import CostMetadata


def calculate_cost(
    metadata: CostMetadata,
    model: str,
) -> float | None:
    """Calculate the cost of an xAI Grok API call.

    https://docs.x.ai

    Pricing (per 1M tokens):

    Model                     Input              Cached             Output
    grok-3                    $3.50              $0.875             $10.50
    grok-3-latest             $3.50              $0.875             $10.50
    grok-2                    $2.00              $0.50              $6.00
    grok-2-latest             $2.00              $0.50              $6.00
    grok-2-1212               $2.00              $0.50              $6.00
    grok-2-mini               $0.33              $0.083             $1.00
    grok-2-vision-1212        $2.00              $0.50              $6.00
    grok-vision-beta          $5.00              $1.25              $15.00
    grok-beta                 $5.00              $1.25              $15.00

    Args:
        input_tokens: Number of input tokens
        cached_tokens: Number of cached tokens
        output_tokens: Number of output tokens
        model: Model name to use for pricing calculation

    Returns:
        Total cost in USD or None if invalid input
    """
    pricing = {
        "grok-3": {
            "prompt": 0.000_003_5,
            "cached": 0.000_000_875,
            "completion": 0.000_010_5,
        },
        "grok-3-latest": {
            "prompt": 0.000_003_5,
            "cached": 0.000_000_875,
            "completion": 0.000_010_5,
        },
        "grok-2": {
            "prompt": 0.000_002,
            "cached": 0.000_000_5,
            "completion": 0.000_006,
        },
        "grok-latest": {
            "prompt": 0.000_002,
            "cached": 0.000_000_5,
            "completion": 0.000_006,
        },
        "grok-2-1212": {
            "prompt": 0.000_002,
            "cached": 0.000_000_5,
            "completion": 0.000_006,
        },
        "grok-2-mini": {
            "prompt": 0.000_000_33,
            "cached": 0.000_000_083,
            "completion": 0.000_001,
        },
        "grok-2-vision-1212": {
            "prompt": 0.000_002,
            "cached": 0.000_000_5,
            "completion": 0.000_006,
        },
        "grok-vision-beta": {
            "prompt": 0.000_005,
            "cached": 0.000_001_25,
            "completion": 0.000_015,
        },
        "grok-beta": {
            "prompt": 0.000_005,
            "cached": 0.000_001_25,
            "completion": 0.000_015,
        },
    }

    if metadata.input_tokens is None or metadata.output_tokens is None:
        return None

    if metadata.cached_tokens is None:
        metadata.cached_tokens = 0

    try:
        model_pricing = pricing[model]
    except KeyError:
        return None

    prompt_price = model_pricing["prompt"]
    cached_price = model_pricing["cached"]
    completion_price = model_pricing["completion"]

    prompt_cost = metadata.input_tokens * prompt_price
    cached_cost = metadata.cached_tokens * cached_price
    completion_cost = metadata.output_tokens * completion_price
    total_cost = prompt_cost + cached_cost + completion_cost

    return total_cost



"""Tests the `_utils.setup_call` function."""

from typing import cast

import pytest

from mirascope.core.base import BaseCallParams, CommonCallParams
from mirascope.core.base._utils._setup_call import setup_call
from mirascope.core.base.dynamic_config import BaseDynamicConfig
from mirascope.core.base.message_param import BaseMessageParam
from mirascope.core.base.prompt import prompt_template
from mirascope.core.base.tool import BaseTool


def test_setup_call() -> None:
    """Tests the `setup_call` function."""

    @prompt_template("Recommend a {genre} book.")
    def fn(genre: str) -> None: ...  # pragma: no cover

    def convert_common_call_params(
        common_params: CommonCallParams,
    ) -> BaseCallParams: ...

    template, messages, tool_types, call_kwargs = setup_call(
        fn,
        {"genre": "fantasy"},
        None,
        None,
        BaseTool,
        {"arg": "value"},  # type: ignore
        convert_common_call_params,  # pyright: ignore [reportArgumentType]
    )

    assert template == "Recommend a {genre} book."
    assert messages == [
        BaseMessageParam(role="user", content="Recommend a fantasy book.")
    ]
    assert tool_types is None
    assert call_kwargs == {"arg": "value"}

    @prompt_template("Recommend a {genre} book.")
    def prompt_template_fn(genre: str) -> None:
        """Normal docstr."""

    assert setup_call(
        prompt_template_fn,
        {"genre": "fantasy"},
        None,
        None,
        BaseTool,
        {"arg": "value"},  # type: ignore
        convert_common_call_params,  # pyright: ignore [reportArgumentType]
    ) == (template, messages, tool_types, call_kwargs)


def test_setup_call_with_dynamic_config() -> None:
    """Tests the `setup_call` function with dynamic config."""

    class FormatBook(BaseTool):
        title: str
        author: str

        def call(self) -> None:
            """Format book tool call method."""

        @classmethod
        def tool_schema(cls):
            return {"type": "function", "name": cls._name()}

    def format_book(title: str, author: str) -> None:
        """Format book tool."""

    dynamic_config: BaseDynamicConfig = {
        "metadata": {"tags": {"version:0001"}},
        "computed_fields": {"genre": "fantasy"},
        "tools": [FormatBook, format_book],
        "call_params": {"arg": "value"},
    }

    @prompt_template("Recommend a {genre} book.")
    def fn() -> None: ...  # pragma: no cover

    def convert_common_call_params(common_params: CommonCallParams) -> BaseCallParams:
        """Test conversion function for common parameters."""
        return cast(BaseCallParams, common_params)

    template, messages, tool_types, call_kwargs = setup_call(
        fn,
        {},
        dynamic_config,
        None,
        FormatBook,
        {},
        convert_common_call_params,  # pyright: ignore [reportArgumentType]
    )

    assert template == "Recommend a {genre} book."
    assert messages == [
        BaseMessageParam(role="user", content="Recommend a fantasy book.")
    ]
    assert tool_types and len(tool_types) == 2
    tool0, tool1 = tool_types
    assert tool0._name() == "FormatBook"
    assert tool1._name() == "format_book"
    assert call_kwargs == {
        "arg": "value",
        "tools": [
            {"type": "function", "name": "FormatBook"},
            {"type": "function", "name": "format_book"},
        ],
    }


@pytest.mark.parametrize("messages_type", [list, tuple])
def test_setup_call_with_custom_messages(messages_type) -> None:
    """Tests the `setup_call` function with custom messages."""

    custom_messages = messages_type(
        {
            "role": "user",
            "content": [{"type": "text", "text": "Recommend a fantasy book."}],
        }
    )
    dynamic_config: BaseDynamicConfig = {"messages": custom_messages}

    def fn() -> None:
        """Normal docstr."""

    def convert_common_call_params(common_params: CommonCallParams) -> BaseCallParams:
        """Test conversion function for common parameters."""
        return cast(BaseCallParams, common_params)

    template, messages, tool_types, call_kwargs = setup_call(
        fn,
        {},
        dynamic_config,
        None,
        BaseTool,
        {},
        convert_common_call_params,  # pyright: ignore [reportArgumentType]
    )
    assert template is None
    assert list(custom_messages) == messages
    assert tool_types is None
    assert call_kwargs == {}


# tests/core/base/_utils/test_setup_call.py


def test_setup_call_with_common_params() -> None:
    """Tests setup_call with CommonCallParams.

    Verifies that CommonCallParams are correctly converted using the provided
    conversion function.
    """
    common_params: CommonCallParams = {
        "temperature": 0.7,
        "max_tokens": 100,
        "top_p": 0.9,
    }

    def convert_common_call_params(params: CommonCallParams) -> BaseCallParams:
        """Test conversion function that prefixes parameters with 'converted_'."""
        return cast(
            BaseCallParams,
            {
                "converted_temperature": params.get("temperature"),
                "converted_max_tokens": params.get("max_tokens"),
                "converted_top_p": params.get("top_p"),
            },
        )

    @prompt_template("Recommend a {genre} book.")
    def fn(genre: str) -> None: ...  # pragma: no cover

    template, messages, tool_types, call_kwargs = setup_call(
        fn,
        {"genre": "fantasy"},
        None,
        None,
        BaseTool,
        common_params,
        convert_common_call_params,  # pyright: ignore [reportArgumentType]
    )

    assert template == "Recommend a {genre} book."
    assert messages == [
        BaseMessageParam(role="user", content="Recommend a fantasy book.")
    ]
    assert tool_types is None
    assert call_kwargs == {
        "converted_temperature": 0.7,
        "converted_max_tokens": 100,
        "converted_top_p": 0.9,
    }


def test_setup_call_with_mixed_common_and_call_params() -> None:
    """Tests setup_call with a mix of CommonCallParams and provider-specific params.

    Verifies that when the input contains both common parameters and provider-specific
    parameters, only the common parameters are converted while others remain unchanged.
    """
    # Dictionary containing both CommonCallParams keys and provider-specific keys
    mixed_params = {
        "temperature": 0.7,  # CommonCallParams key
        "max_tokens": 100,  # CommonCallParams key
        "custom_param": "value",  # Provider-specific key
    }

    def convert_common_call_params(params: CommonCallParams) -> BaseCallParams: ...

    @prompt_template("Recommend a {genre} book.")
    def fn(genre: str) -> None: ...  # pragma: no cover

    template, messages, tool_types, call_kwargs = setup_call(
        fn,
        {"genre": "fantasy"},
        None,
        None,
        BaseTool,
        mixed_params,  # type: ignore
        convert_common_call_params,  # pyright: ignore [reportArgumentType]
    )

    # Provider-specific parameters remain unchanged while common parameters are converted
    assert call_kwargs == mixed_params


def test_setup_call_with_dynamic_config_and_common_params() -> None:
    """Tests setup_call with both dynamic config and CommonCallParams.

    Verifies that parameters from dynamic config are properly merged with
    converted common parameters.
    """
    common_params: CommonCallParams = {
        "temperature": 0.7,
        "max_tokens": 100,
    }

    dynamic_config: BaseDynamicConfig = {
        "call_params": {
            "custom_param": "value",
        }
    }

    def convert_common_call_params(params: CommonCallParams) -> BaseCallParams:
        """Test conversion function for common parameters."""
        return cast(
            BaseCallParams,
            {
                "converted_temperature": params.get("temperature"),
                "converted_max_tokens": params.get("max_tokens"),
            },
        )

    @prompt_template("Recommend a {genre} book.")
    def fn(genre: str) -> None: ...  # pragma: no cover

    template, messages, tool_types, call_kwargs = setup_call(
        fn,
        {"genre": "fantasy"},
        dynamic_config,
        None,
        BaseTool,
        common_params,
        convert_common_call_params,  # pyright: ignore [reportArgumentType]
    )

    # Results from convert_common_call_params are merged with dynamic_config's call_params
    assert call_kwargs == {
        "converted_temperature": 0.7,
        "converted_max_tokens": 100,
        "custom_param": "value",
    }


def test_setup_call_with_empty_common_params() -> None:
    """Tests setup_call with empty CommonCallParams.

    Verifies that the function handles empty parameter dictionaries correctly
    and maintains other expected behavior.
    """

    def convert_common_call_params(params: CommonCallParams) -> BaseCallParams:
        """Test conversion function for common parameters."""
        return cast(BaseCallParams, params)

    @prompt_template("Recommend a {genre} book.")
    def fn(genre: str) -> None: ...  # pragma: no cover

    template, messages, tool_types, call_kwargs = setup_call(
        fn,
        {"genre": "fantasy"},
        None,
        None,
        BaseTool,
        {},  # Empty CommonCallParams
        convert_common_call_params,  # pyright: ignore [reportArgumentType]
    )

    assert template == "Recommend a {genre} book."
    assert messages == [
        BaseMessageParam(role="user", content="Recommend a fantasy book.")
    ]
    assert tool_types is None
    assert call_kwargs == {}



"""Tests the `bedrock.stream` module."""

from collections.abc import Generator

import pytest
from mypy_boto3_bedrock_runtime.type_defs import (
    ConverseStreamMetadataEventTypeDef,
    ResponseMetadataTypeDef,
    ToolTypeDef,
)

from mirascope.core.base.types import CostMetadata
from mirascope.core.bedrock.call_params import BedrockCallParams
from mirascope.core.bedrock.call_response import BedrockCallResponse
from mirascope.core.bedrock.call_response_chunk import BedrockCallResponseChunk
from mirascope.core.bedrock.stream import BedrockStream
from mirascope.core.bedrock.tool import BedrockTool


class MockTool(BedrockTool):
    name: str

    def call(self) -> str: ...


def test_bedrock_stream_init():
    stream = BedrockStream(
        stream=iter([]),
        metadata={},
        tool_types=None,
        call_response_type=BedrockCallResponse,
        model="anthropic.claude-v2",
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params=BedrockCallParams(),
        call_kwargs={},
    )
    assert stream._provider == "bedrock"
    assert stream._metadata is None
    assert stream._response_metadata == {
        "RequestId": "",
        "HTTPStatusCode": 500,
        "HTTPHeaders": {},
        "RetryAttempts": 0,
    }
    assert stream.cost is None
    assert stream.cost_metadata == CostMetadata()


def test_bedrock_stream_construct_message_param():
    stream = BedrockStream(
        stream=iter([]),
        metadata={},
        tool_types=None,
        call_response_type=BedrockCallResponse,
        model="anthropic.claude-v2",
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params=BedrockCallParams(),
        call_kwargs={},
    )

    # Test with content only
    message_param = stream._construct_message_param(content="Test content")
    assert message_param == {"role": "assistant", "content": [{"text": "Test content"}]}

    # Test with tool calls only
    tool_calls = [ToolTypeDef(toolUseId="1", name="TestTool", input={})]  # pyright: ignore [reportCallIssue]
    message_param = stream._construct_message_param(tool_calls=tool_calls)  # pyright: ignore [reportArgumentType]
    assert message_param == {
        "role": "assistant",
        "content": [{"toolUse": tool_calls[0]}],
    }

    # Test with both content and tool calls
    message_param = stream._construct_message_param(
        content="Test content",
        tool_calls=tool_calls,  # pyright: ignore [reportArgumentType]
    )
    assert message_param == {
        "role": "assistant",
        "content": [{"text": "Test content"}, {"toolUse": tool_calls[0]}],
    }


def test_bedrock_stream_update_properties():
    stream = BedrockStream(
        stream=iter([]),
        metadata={},
        tool_types=None,
        call_response_type=BedrockCallResponse,
        model="anthropic.claude-v2",
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params=BedrockCallParams(),
        call_kwargs={},
    )

    chunk = BedrockCallResponseChunk(
        chunk={  # pyright: ignore [reportArgumentType]
            "model": "anthropic.claude-v2",
            "metadata": ConverseStreamMetadataEventTypeDef(  # pyright: ignore [reportCallIssue]
                usage={"inputTokens": 10, "outputTokens": 20}
            ),
            "responseMetadata": ResponseMetadataTypeDef(
                RequestId="test-id", HTTPStatusCode=200, HTTPHeaders={}, RetryAttempts=0
            ),
        }
    )

    stream._update_properties(chunk)
    assert stream._metadata == chunk.chunk["metadata"]  # pyright: ignore [reportTypedDictNotRequiredAccess]
    assert stream._response_metadata == chunk.chunk["responseMetadata"]


def test_bedrock_stream_construct_call_response():
    stream = BedrockStream(
        stream=iter([]),
        metadata={},
        tool_types=[MockTool],
        call_response_type=BedrockCallResponse,
        model="anthropic.claude-v2",
        prompt_template="Test prompt",
        fn_args={"arg": "value"},
        dynamic_config=None,
        messages=[],
        call_params=BedrockCallParams(),
        call_kwargs={},
    )

    stream.message_param = {"role": "assistant", "content": [{"text": "Test content"}]}
    stream.finish_reasons = ["end_turn"]
    stream._metadata = ConverseStreamMetadataEventTypeDef(
        usage={"inputTokens": 10, "outputTokens": 20},  # pyright: ignore [reportArgumentType]
        metrics={"latencyMs": 100},
        trace={"id": "trace-id"},  # pyright: ignore [reportArgumentType]
    )
    stream._response_metadata = ResponseMetadataTypeDef(
        RequestId="test-id", HTTPStatusCode=200, HTTPHeaders={}, RetryAttempts=0
    )
    stream.start_time = 0
    stream.end_time = 1

    response = stream.construct_call_response()
    assert isinstance(response, BedrockCallResponse)
    assert response.content == "Test content"
    assert response.finish_reasons == ["end_turn"]
    assert response.tool_types == [MockTool]
    assert response.prompt_template == "Test prompt"
    assert response.fn_args == {"arg": "value"}


def test_bedrock_stream_construct_call_response_error():
    stream = BedrockStream(
        stream=iter([]),
        metadata={},
        tool_types=None,
        call_response_type=BedrockCallResponse,
        model="anthropic.claude-v2",
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params=BedrockCallParams(),
        call_kwargs={},
    )

    with pytest.raises(
        ValueError, match="No stream response, check if the stream has been consumed."
    ):
        stream.construct_call_response()


def test_bedrock_stream_iteration():
    response_metadata = ResponseMetadataTypeDef(
        RequestId="test-id", HTTPStatusCode=200, HTTPHeaders={}, RetryAttempts=0
    )
    chunks = [
        BedrockCallResponseChunk(
            chunk={  # pyright: ignore [reportArgumentType]
                "model": "anthropic.claude-v2",
                "responseMetadata": response_metadata,
                "contentBlockDelta": {"delta": {"text": "Hello"}},
            }
        ),
        BedrockCallResponseChunk(
            chunk={  # pyright: ignore [reportArgumentType]
                "model": "anthropic.claude-v2",
                "responseMetadata": response_metadata,
                "contentBlockDelta": {"delta": {"text": " world"}},
            }
        ),
        BedrockCallResponseChunk(
            chunk={
                "model": "anthropic.claude-v2",
                "responseMetadata": response_metadata,
                "messageStop": {"stopReason": "end_turn"},
            }
        ),
    ]

    def chunk_generator() -> Generator[
        tuple[BedrockCallResponseChunk, None], None, None
    ]:
        for chunk in chunks:
            yield chunk, None

    stream = BedrockStream(
        stream=chunk_generator(),
        metadata={},
        tool_types=None,
        call_response_type=BedrockCallResponse,
        model="anthropic.claude-v2",
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params=BedrockCallParams(),
        call_kwargs={},
    )

    collected_chunks = list(stream)
    assert len(collected_chunks) == 3
    assert collected_chunks[0][0].content == "Hello"
    assert collected_chunks[1][0].content == " world"
    assert collected_chunks[2][0].finish_reasons == ["end_turn"]
    assert stream.message_param == {
        "role": "assistant",
        "content": [{"text": "Hello world"}],
    }



from unittest.mock import MagicMock, patch

from mirascope.tools import Requests, RequestsConfig


def test_requests_config():
    config = RequestsConfig()
    assert config.timeout == 5

    custom_config = RequestsConfig(timeout=10)
    assert custom_config.timeout == 10


@patch("mirascope.tools.web._requests.requests")
def test_requests_get_success(mock_request):
    mock_response = MagicMock()
    mock_response.text = "Test content"
    mock_request.request.return_value = mock_response

    tool = Requests(url="https://example.com")  # pyright: ignore [reportCallIssue]
    result = tool.call()

    assert result == "Test content"
    mock_request.request.assert_called_with(
        method="GET",
        url="https://example.com",
        json=None,
        headers=None,
        timeout=tool._get_config().timeout,
    )


@patch("mirascope.tools.web._requests.requests")
def test_requests_post_with_data(mock_request):
    mock_response = MagicMock()
    mock_response.text = "Test response"
    mock_request.request.return_value = mock_response

    tool = Requests(
        url="https://example.com",
        method="POST",
        data={"key": "value"},
        headers={"Content-Type": "application/json"},
    )
    result = tool.call()

    assert result == "Test response"
    mock_request.request.assert_called_with(
        method="POST",
        url="https://example.com",
        json={"key": "value"},
        headers={"Content-Type": "application/json"},
        timeout=tool._get_config().timeout,
    )


@patch("mirascope.tools.web._requests.requests")
def test_requests_error(mock_request):
    mock_request.request.side_effect = Exception("Request failed")

    tool = Requests(url="https://example.com")  # pyright: ignore [reportCallIssue]
    result = tool.call()
    assert "Failed to extract content from URL" in result



"""Tests the `mistral._utils.handle_stream` module."""

import pytest
from mistralai.models import (
    CompletionChunk,
    CompletionEvent,
    CompletionResponseStreamChoice,
    DeltaMessage,
    FunctionCall,
    ToolCall,
)

from mirascope.core.mistral._utils._handle_stream import (
    handle_stream,
    handle_stream_async,
)
from mirascope.core.mistral.tool import MistralTool


class FormatBook(MistralTool):
    """Returns the title and author nicely formatted."""

    title: str
    author: str

    def call(self) -> None:
        """Dummy call."""


@pytest.fixture()
def mock_chunks() -> list[CompletionChunk]:
    """Returns a list of mock `ChatCompletionStreamResponse` instances."""

    new_tool_call = ToolCall(
        id="id",
        function=FunctionCall(
            arguments="",
            name="FormatBook",
        ),
        type="function",
    )
    tool_call = ToolCall(
        id="null",
        function=FunctionCall(
            arguments='{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
            name="FormatBook",
        ),
        type="function",
    )
    return [
        CompletionChunk(
            id="id",
            choices=[
                CompletionResponseStreamChoice(
                    index=0,
                    delta=DeltaMessage(content="content", tool_calls=None),
                    finish_reason=None,
                )
            ],
            created=0,
            model="mistral-large-latest",
            object="chat.completion.chunk",
        ),
        CompletionChunk(
            id="id",
            choices=[
                CompletionResponseStreamChoice(
                    index=0,
                    delta=DeltaMessage(
                        content=None,
                        tool_calls=[new_tool_call],
                    ),
                    finish_reason=None,
                )
            ],
            created=0,
            model="mistral-large-latest",
            object="chat.completion.chunk",
        ),
        CompletionChunk(
            id="id",
            choices=[
                CompletionResponseStreamChoice(
                    index=0,
                    delta=DeltaMessage(
                        content=None,
                        tool_calls=[tool_call],
                    ),
                    finish_reason=None,
                )
            ],
            created=0,
            model="mistral-large-latest",
            object="chat.completion.chunk",
        ),
        CompletionChunk(
            id="id",
            choices=[
                CompletionResponseStreamChoice(
                    index=0,
                    delta=DeltaMessage(
                        content=None,
                        tool_calls=[new_tool_call],
                    ),
                    finish_reason=None,
                )
            ],
            created=0,
            model="mistral-large-latest",
            object="chat.completion.chunk",
        ),
        CompletionChunk(
            id="id",
            choices=[
                CompletionResponseStreamChoice(
                    index=0,
                    delta=DeltaMessage(
                        content=None,
                        tool_calls=[tool_call],
                    ),
                    finish_reason=None,
                )
            ],
            created=0,
            model="mistral-large-latest",
            object="chat.completion.chunk",
        ),
        CompletionChunk(
            id="id",
            choices=[
                CompletionResponseStreamChoice(
                    index=0,
                    delta=DeltaMessage(content=None, tool_calls=None),
                    finish_reason="tool_calls",
                )
            ],
            created=0,
            model="mistral-large-latest",
            object="chat.completion.chunk",
        ),
    ]


@pytest.fixture()
def mock_chunks_onetime_tools() -> list[CompletionChunk]:
    """Returns a list of mock `ChatCompletionStreamResponse` instances."""

    tool_call = ToolCall(
        id="id",
        function=FunctionCall(
            arguments='{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
            name="FormatBook",
        ),
        type="function",
    )
    return [
        CompletionChunk(
            id="id",
            choices=[
                CompletionResponseStreamChoice(
                    index=0,
                    delta=DeltaMessage(content="content", tool_calls=None),
                    finish_reason=None,
                )
            ],
            created=0,
            model="mistral-large-latest",
            object="chat.completion.chunk",
        ),
        CompletionChunk(
            id="id",
            choices=[
                CompletionResponseStreamChoice(
                    index=0,
                    delta=DeltaMessage(
                        content=None,
                        tool_calls=[tool_call],
                    ),
                    finish_reason=None,
                )
            ],
            created=0,
            model="mistral-large-latest",
            object="chat.completion.chunk",
        ),
    ]


def test_handle_stream(mock_chunks: list[CompletionChunk]) -> None:
    """Tests the `handle_stream` function."""

    result = list(
        handle_stream(
            (CompletionEvent(data=c) for c in mock_chunks), tool_types=[FormatBook]
        )
    )
    # Check we get three tuples back.
    # (chunk, None), (chunk, FormatBook), (chunk, FormatBook)
    assert len(result) == 3
    assert result[0][1] is None
    assert (
        (tool := result[1][1]) is not None
        and isinstance(tool, FormatBook)
        and tool.model_dump(exclude={"tool_call", "delta"})
        == {"title": "The Name of the Wind", "author": "Patrick Rothfuss"}
    )
    assert (
        (tool := result[2][1]) is not None
        and isinstance(tool, FormatBook)
        and tool.model_dump(exclude={"tool_call", "delta"})
        == {"title": "The Name of the Wind", "author": "Patrick Rothfuss"}
    )


@pytest.mark.asyncio
async def test_handle_stream_async(
    mock_chunks: list[CompletionChunk],
) -> None:
    """Tests the `handle_stream_async` function."""

    async def generator():
        for chunk in mock_chunks:
            yield CompletionEvent(data=chunk)

    result = []
    async for t in handle_stream_async(generator(), tool_types=[FormatBook]):
        result.append(t)
    # Check we get three tuples back.
    # (chunk, None), (chunk, FormatBook), (chunk, FormatBook)
    assert len(result) == 3
    assert result[0][1] is None
    assert (
        (tool := result[1][1]) is not None
        and isinstance(tool, FormatBook)
        and tool.model_dump(exclude={"tool_call", "delta"})
        == {"title": "The Name of the Wind", "author": "Patrick Rothfuss"}
    )
    assert (
        (tool := result[2][1]) is not None
        and isinstance(tool, FormatBook)
        and tool.model_dump(exclude={"tool_call", "delta"})
        == {"title": "The Name of the Wind", "author": "Patrick Rothfuss"}
    )


def test_handle_stream_onetime_tools(mock_chunks_onetime_tools) -> None:
    """Tests the `handle_stream` function."""

    result = list(
        handle_stream(
            (CompletionEvent(data=c) for c in mock_chunks_onetime_tools),
            tool_types=[FormatBook],
        )
    )
    # Check we get three tuples back.
    # (chunk, None), (chunk, FormatBook), (chunk, FormatBook)
    assert len(result) == 2
    assert result[0][1] is None
    assert (
        (tool := result[1][1]) is not None
        and isinstance(tool, FormatBook)
        and tool.model_dump(exclude={"tool_call", "delta"})
        == {"title": "The Name of the Wind", "author": "Patrick Rothfuss"}
    )


@pytest.mark.asyncio
async def test_handle_stream_async_onetime_tools(
    mock_chunks_onetime_tools,
) -> None:
    """Tests the `handle_stream_async` function."""

    async def generator():
        for chunk in mock_chunks_onetime_tools:
            yield CompletionEvent(data=chunk)

    result = []
    async for t in handle_stream_async(generator(), tool_types=[FormatBook]):
        result.append(t)
    # Check we get three tuples back.
    # (chunk, None), (chunk, FormatBook), (chunk, FormatBook)
    assert len(result) == 2
    assert result[0][1] is None
    assert (
        (tool := result[1][1]) is not None
        and isinstance(tool, FormatBook)
        and tool.model_dump(exclude={"tool_call", "delta"})
        == {"title": "The Name of the Wind", "author": "Patrick Rothfuss"}
    )



"""Tests the `vertex.call_response` module."""

from google.cloud.aiplatform_v1beta1.types import (
    FunctionCall,
    FunctionResponse,
    GenerateContentResponse,
)
from vertexai.generative_models import (
    Candidate,
    Content,
    Part,
)
from vertexai.generative_models import (  # type: ignore
    GenerationResponse as GenerateContentResponseType,
)

from mirascope.core import BaseMessageParam
from mirascope.core.vertex.call_response import VertexCallResponse
from mirascope.core.vertex.tool import VertexTool


def test_vertex_call_response() -> None:
    """Tests the `VertexCallResponse` class."""
    response = GenerateContentResponseType.from_dict(
        {
            "candidates": [
                Candidate.from_dict(
                    {
                        "finish_reason": 1,
                        "content": Content(
                            parts=[Part.from_text("The author is Patrick Rothfuss")],
                            role="model",
                        ).to_dict(),
                    }
                ).to_dict()
            ],
            "usage_metadata": {
                "prompt_token_count": 10,
                "candidates_token_count": 20,
                "total_token_count": 30,
            },
        }
    )
    call_response = VertexCallResponse(
        metadata={},
        response=response,
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    call_response._model = "gemini-1.5-flash"
    assert call_response._provider == "vertex"
    assert call_response.content == "The author is Patrick Rothfuss"
    assert call_response.finish_reasons == ["STOP"]
    assert call_response.model == "gemini-1.5-flash"
    assert call_response.id is None
    assert call_response.tools is None
    assert call_response.tool is None
    assert isinstance(call_response.usage, GenerateContentResponse.UsageMetadata)
    assert call_response.usage.prompt_token_count == 10
    assert call_response.usage.candidates_token_count == 20
    assert call_response.usage.total_token_count == 30
    assert call_response.output_tokens == 20
    assert call_response.cost == 1.6874999999999997e-06
    assert call_response.message_param.to_dict() == {
        "parts": [{"text": "The author is Patrick Rothfuss"}],
        "role": "model",
    }
    assert call_response.common_finish_reasons == ["stop"]
    assert call_response.common_message_param == BaseMessageParam(
        role="assistant",
        content="The author is Patrick Rothfuss",
    )
    assert call_response.common_user_message_param is None
    call_response.user_message_param = Content(
        parts=[Part.from_text("content")], role="user"
    )
    assert call_response.common_user_message_param == BaseMessageParam(
        role="user", content="content"
    )


def test_vertex_call_response_with_tools() -> None:
    """Tests the `VertexCallResponse` class with tools."""

    class FormatBook(VertexTool):
        title: str
        author: str

        def call(self) -> str:
            return f"{self.title} by {self.author}"

    raw_part = Part()
    raw_part._raw_part.function_call = FunctionCall(
        name="FormatBook",
        args={
            "title": "The Name of the Wind",
            "author": "Patrick Rothfuss",
        },
    )
    response = GenerateContentResponseType.from_dict(
        {
            "candidates": [
                Candidate.from_dict(
                    {
                        "finishReason": 1,
                        "content": Content(parts=[raw_part]).to_dict(),
                    },
                ).to_dict()
            ]
        }
    )
    call_response = VertexCallResponse(
        metadata={},
        response=response,
        tool_types=[FormatBook],
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    tools = call_response.tools
    tool = call_response.tool
    assert tools and len(tools) == 1 and tools[0] == tool
    assert isinstance(tool, FormatBook)
    assert tool.title == "The Name of the Wind"
    assert tool.author == "Patrick Rothfuss"
    output = tool.call()
    assert output == "The Name of the Wind by Patrick Rothfuss"
    tool_message_params = call_response.tool_message_params([(tool, output)])
    assert len(tool_message_params) == 1
    assert tool_message_params[0].role == "user"
    part = tool_message_params[0].parts[0]
    assert part.function_response == FunctionResponse(
        name="FormatBook", response={"result": output}
    )



from unittest.mock import MagicMock, patch

import pytest

from mirascope.core.base import (
    BaseCallResponse,
    CommonCallParams,
)
from mirascope.core.openai import OpenAICallParams
from mirascope.llm._override import override


class ConcreteMockResponse(BaseCallResponse):
    @property
    def content(self): ...  # pyright: ignore [reportIncompatibleMethodOverride]

    @property
    def finish_reasons(self): ...

    @property
    def model(self): ...

    @property
    def id(self): ...

    @property
    def usage(self): ...

    @property
    def input_tokens(self): ...

    @property
    def output_tokens(self): ...

    @property
    def cost(self): ...

    @property
    def message_param(self): ...  # pyright: ignore [reportIncompatibleVariableOverride]

    @property
    def tools(self): ...  # pyright: ignore [reportIncompatibleVariableOverride]

    @property
    def tool(self): ...  # pyright: ignore [reportIncompatibleVariableOverride]

    @classmethod
    def tool_message_params(cls, tools_and_outputs): ...  # pyright: ignore [reportIncompatibleMethodOverride]

    @property
    def common_finish_reasons(self): ...

    @property
    def common_message_param(self): ...  # pyright: ignore [reportIncompatibleMethodOverride]

    @property
    def common_tools(self): ...

    @property
    def common_usage(self): ...

    def common_construct_call_response(self): ...

    def common_construct_message_param(self, tool_calls, content): ...


def test_override_error_if_only_provider():
    with pytest.raises(
        ValueError,
        match="Provider and model must both be overridden if either is overridden.",
    ):
        override(  # pyright: ignore [reportCallIssue]
            provider_agnostic_call=MagicMock(),
            provider="anthropic",
            model=None,  # pyright: ignore [reportArgumentType]
            call_params=None,
        )


def test_override_with_model():
    with patch("mirascope.llm._override._context") as mock_context:
        fn = override(
            provider_agnostic_call=MagicMock(),
            provider="openai",
            model="overridden-model",
            call_params=CommonCallParams(),
        )
        fn()
        mock_context.assert_called_once()
        _, kwargs = mock_context.call_args
        assert kwargs["provider"] == "openai"
        assert kwargs["model"] == "overridden-model"


def test_override_with_callparams_override():
    new_call_params = OpenAICallParams(temperature=0.7, max_tokens=1111)
    with patch("mirascope.llm._override._context") as mock_context:
        fn = override(  # pyright: ignore [reportCallIssue]
            provider_agnostic_call=MagicMock(),
            provider="openai",
            model="gpt-4o-mini",  # pyright: ignore [reportArgumentType]
            call_params=new_call_params,
        )
        fn()
        mock_context.assert_called_once()
        _, kwargs = mock_context.call_args
        assert kwargs["call_params"] is new_call_params


def test_override_with_client():
    new_client = object()
    with patch("mirascope.llm._override._context") as mock_context:
        fn = override(
            provider_agnostic_call=MagicMock(),
            provider="openai",
            model="overridden-model",
            call_params=CommonCallParams(),
            client=new_client,
        )
        fn()
        mock_context.assert_called_once()
        _, kwargs = mock_context.call_args
        assert kwargs["client"] is new_client


@pytest.mark.asyncio
async def test_override_with_async_gather():
    """Test that override works correctly with asyncio.gather."""
    import asyncio

    # Mock the provider_agnostic_call
    async def mock_async_fn():
        return MagicMock()

    # Mock the _context to capture the context parameters
    with patch("mirascope.llm._override._context") as mock_context:
        # Set up the mock to record calls and return a context manager
        context_manager = MagicMock()
        context_manager.__enter__ = MagicMock(return_value=None)
        context_manager.__exit__ = MagicMock(return_value=None)
        mock_context.return_value = context_manager

        # Create two overridden functions with different providers/models
        openai_fn = override(
            provider_agnostic_call=mock_async_fn,
            provider="openai",
            model="gpt-4o-mini",
            call_params=None,
        )

        anthropic_fn = override(
            provider_agnostic_call=mock_async_fn,
            provider="anthropic",
            model="claude-3-5-sonnet",
            call_params=None,
        )

        # Create futures for both functions
        openai_future = openai_fn()
        anthropic_future = anthropic_fn()

        # Await both futures together
        await asyncio.gather(openai_future, anthropic_future)

        # Check that both contexts were created with the correct parameters
        assert mock_context.call_count == 2

        # Extract the call arguments
        call_args_list = mock_context.call_args_list

        # First call should be for openai
        _, kwargs1 = call_args_list[0]
        assert kwargs1["provider"] == "openai"
        assert kwargs1["model"] == "gpt-4o-mini"

        # Second call should be for anthropic
        _, kwargs2 = call_args_list[1]
        assert kwargs2["provider"] == "anthropic"
        assert kwargs2["model"] == "claude-3-5-sonnet"



"""Tests the `openai._utils.convert_message_params` function."""

import base64
from unittest.mock import patch

import pytest
from openai.types.chat import ChatCompletionMessageParam

from mirascope.core.base import (
    AudioPart,
    AudioURLPart,
    BaseMessageParam,
    CacheControlPart,
    ImagePart,
    ImageURLPart,
    TextPart,
    ToolCallPart,
    ToolResultPart,
)
from mirascope.core.openai._utils._convert_message_params import convert_message_params


def test_convert_message_params() -> None:
    """Tests the `convert_message_params` function."""

    message_params: list[BaseMessageParam | ChatCompletionMessageParam] = [
        {"role": "user", "content": [{"type": "text", "text": "Hello"}]},
        BaseMessageParam(role="user", content="Hello"),
        BaseMessageParam(
            role="user",
            content=[
                TextPart(type="text", text="Hello"),
                ImagePart(
                    type="image", media_type="image/jpeg", image=b"image", detail="auto"
                ),
                ImageURLPart(
                    type="image_url", url="http://example.com/image", detail="auto"
                ),
                AudioPart(type="audio", media_type="audio/wav", audio=b"audio"),
            ],
        ),
        BaseMessageParam(
            role="assistant",
            content=[
                ToolCallPart(type="tool_call", name="tool_name", id="tool_id"),
            ],
        ),
        BaseMessageParam(
            role="user",
            content=[
                TextPart(type="text", text="Here you go!"),
                ToolResultPart(
                    name="tool_name", id="tool_id", content="result", type="tool_result"
                ),
            ],
        ),
    ]
    converted_message_params = convert_message_params(message_params)
    assert converted_message_params == [
        {"content": [{"text": "Hello", "type": "text"}], "role": "user"},
        {"content": "Hello", "role": "user"},
        {
            "content": [
                {"text": "Hello", "type": "text"},
                {
                    "image_url": {
                        "detail": "auto",
                        "url": "data:image/jpeg;base64,aW1hZ2U=",
                    },
                    "type": "image_url",
                },
                {
                    "image_url": {"detail": "auto", "url": "http://example.com/image"},
                    "type": "image_url",
                },
                {
                    "input_audio": {"data": "YXVkaW8=", "format": "wav"},
                    "type": "input_audio",
                },
            ],
            "role": "user",
        },
        {
            "role": "assistant",
            "tool_calls": [
                {
                    "function": {"arguments": "null", "name": "tool_name"},
                    "type": "function",
                    "id": "tool_id",
                }
            ],
        },
        {"content": [{"text": "Here you go!", "type": "text"}], "role": "user"},
        {"content": "result", "role": "tool", "tool_call_id": "tool_id"},
    ]

    with pytest.raises(
        ValueError,
        match="Unsupported image media type: image/svg. OpenAI currently only supports "
        "JPEG, PNG, GIF, and WebP images.",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        ImagePart(
                            type="image",
                            media_type="image/svg",
                            image=b"image",
                            detail="auto",
                        )
                    ],
                )
            ]
        )

    with pytest.raises(
        ValueError,
        match="Unsupported audio media type: audio/aac. OpenAI currently only supports WAV and MP3 audio file types.",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        AudioPart(type="audio", media_type="audio/aac", audio=b"audio")
                    ],
                )
            ]
        )

    with pytest.raises(
        ValueError,
        match="OpenAI currently only supports text, image and audio parts. Part provided: cache_control",
    ):
        convert_message_params(
            [
                BaseMessageParam(
                    role="user",
                    content=[
                        CacheControlPart(type="cache_control", cache_type="ephemeral")
                    ],
                )
            ]
        )


@patch(
    "mirascope.core.openai._utils._convert_message_params._load_media",
    return_value=b"audio_data",
)
@patch(
    "mirascope.core.openai._utils._convert_message_params.get_audio_type",
    return_value="wav",
)
def test_audio_url_valid(mock_get_audio_type, mock_load_media) -> None:
    """Tests conversion of a valid audio_url part."""
    message = BaseMessageParam(
        role="user",
        content=[
            AudioURLPart(
                type="audio_url",
                url="http://example.com/audio",
                detail="ignored",  # pyright: ignore [reportCallIssue]
            )
        ],
    )
    result = convert_message_params([message])
    expected = {
        "content": [
            {
                "input_audio": {
                    "format": "wav",
                    "data": base64.b64encode(b"audio_data").decode("utf-8"),
                },
                "type": "input_audio",
            }
        ],
        "role": "user",
    }
    assert result == [expected]
    mock_load_media.assert_called_once_with("http://example.com/audio")
    mock_get_audio_type.assert_called_once_with(b"audio_data")


@patch(
    "mirascope.core.openai._utils._convert_message_params._load_media",
    return_value=b"audio_data",
)
@patch(
    "mirascope.core.openai._utils._convert_message_params.get_audio_type",
    return_value="aiff",
)
def test_audio_url_invalid(mock_get_audio_type, mock_load_media) -> None:
    """Tests conversion of an audio_url part with an unsupported audio type."""
    message = BaseMessageParam(
        role="user",
        content=[
            AudioURLPart(
                type="audio_url",
                url="http://example.com/audio",
                detail="ignored",  # pyright: ignore [reportCallIssue]
            )
        ],
    )
    with pytest.raises(
        ValueError,
        match="Unsupported audio media type: audio/aiff. OpenAI currently only supports WAV and MP3 audio file types.",
    ):
        convert_message_params([message])
    mock_load_media.assert_called_once_with("http://example.com/audio")
    mock_get_audio_type.assert_called_once_with(b"audio_data")



"""Tests the `google._utils.setup_call` module."""

from collections.abc import Callable
from typing import cast
from unittest.mock import MagicMock, patch

import pytest
from google.genai.types import (
    FunctionCallingConfig,
    FunctionCallingConfigMode,
    GenerateContentConfig,
    Part,
    Tool,
)
from pydantic import BaseModel

from mirascope.core.google._utils._convert_common_call_params import (
    convert_common_call_params,
)
from mirascope.core.google._utils._setup_call import setup_call
from mirascope.core.google.call_params import GoogleCallParams
from mirascope.core.google.tool import GoogleTool


@pytest.fixture()
def mock_base_setup_call() -> MagicMock:
    """Returns the mock setup_call function."""
    mock_setup_call = MagicMock()
    mock_setup_call.return_value = [MagicMock() for _ in range(3)] + [{}]
    return mock_setup_call


@patch(
    "mirascope.core.google._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.google._utils._setup_call._utils", new_callable=MagicMock)
@patch("mirascope.core.google._utils._setup_call.Client", new_callable=MagicMock)
def test_setup_call(
    mock_client_class: MagicMock,
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    """Tests the `setup_call` function."""
    mock_client = MagicMock()
    mock_client_class.return_value = mock_client
    mock_utils.setup_call = mock_base_setup_call
    fn = MagicMock()
    create, prompt_template, messages, tool_types, call_kwargs = setup_call(
        model="google-1.5-flash",
        client=None,
        fn=fn,
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        response_model=None,
        stream=False,
    )
    assert prompt_template == mock_base_setup_call.return_value[0]
    assert tool_types == mock_base_setup_call.return_value[2]
    assert "contents" in call_kwargs and call_kwargs["contents"] == messages
    mock_base_setup_call.assert_called_once_with(
        fn, {}, None, None, GoogleTool, {}, convert_common_call_params
    )
    mock_convert_message_params.assert_called_once_with(
        mock_base_setup_call.return_value[1], mock_client
    )
    assert messages == mock_convert_message_params.return_value
    mock_client_class.assert_called_once_with()
    assert create(**call_kwargs)
    mock_client.models.generate_content.assert_called_once_with(**call_kwargs)
    mock_client.models.generate_content.reset_mock()
    assert create(stream=True, **call_kwargs)
    mock_client.models.generate_content_stream.assert_called_once_with(**call_kwargs)
    mock_client.models.generate_content_stream.reset_mock()
    assert create(stream=False, **call_kwargs)
    mock_client.models.generate_content.assert_called_once_with(**call_kwargs)


@pytest.mark.parametrize(
    "generation_config_type",
    [dict, GenerateContentConfig],
)
@patch(
    "mirascope.core.google._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.google._utils._setup_call._utils", new_callable=MagicMock)
@patch("mirascope.core.google._utils._setup_call.Client", new_callable=MagicMock)
def test_setup_call_json_mode(
    mock_client_class: MagicMock,
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
    generation_config_type: type,
) -> None:
    """Tests the `setup_call` function with JSON mode."""
    mock_client = MagicMock()
    mock_client_class.return_value = mock_client
    mock_utils.setup_call = mock_base_setup_call
    mock_utils.json_mode_content = MagicMock()
    mock_base_setup_call.return_value[1] = [
        {
            "role": "system",
            "parts": [{"text": "this is system message"}],
        },
        {"role": "user", "parts": [{"type": "text", "text": "test"}]},
    ]
    tools: list[Tool | Callable] = [MagicMock(spec=Tool)]
    mock_base_setup_call.return_value[-1]["tools"] = tools
    mock_base_setup_call.return_value[-1]["config"] = generation_config_type(
        candidate_count=1,
        max_output_tokens=100,
        response_mime_type="application/xml",
        response_schema=None,
        stop_sequences=["\n"],
        temperature=0.5,
        top_k=0,
        top_p=0,
    )
    mock_convert_message_params.side_effect = lambda x, y: x
    _, _, messages, _, call_kwargs = setup_call(
        model="google-1.5-flash",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=True,
        call_params={},
        response_model=None,
        stream=False,
    )
    mock_client_class.assert_called_once_with()
    assert isinstance(messages[-1], dict)
    assert "parts" in messages[-1]
    assert isinstance(messages[-1]["parts"], list)
    assert messages[-1]["parts"][-1] == {
        "text": mock_utils.json_mode_content.return_value
    }
    assert "config" in call_kwargs
    assert call_kwargs["config"] == GenerateContentConfig(
        system_instruction=[
            Part(
                video_metadata=None,
                thought=None,
                code_execution_result=None,
                executable_code=None,
                file_data=None,
                function_call=None,
                function_response=None,
                inline_data=None,
                text="this is system message",
            )
        ],
        temperature=0.5,
        top_p=0.0,
        top_k=0.0,
        candidate_count=1,
        max_output_tokens=100,
        stop_sequences=["\n"],
        response_logprobs=None,
        logprobs=None,
        presence_penalty=None,
        frequency_penalty=None,
        seed=None,
        response_mime_type="application/json",
        response_schema=None,
        routing_config=None,
        safety_settings=None,
        tools=tools,
        tool_config=None,
        cached_content=None,
        response_modalities=None,
        media_resolution=None,
        speech_config=None,
        audio_timestamp=None,
        automatic_function_calling=None,
        thinking_config=None,
    )


@patch(
    "mirascope.core.google._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.google._utils._setup_call._utils", new_callable=MagicMock)
@patch("mirascope.core.google._utils._setup_call.Client", new_callable=MagicMock)
def test_setup_call_extract(
    mock_client_class: MagicMock,
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    """Tests the `setup_call` function with extraction."""
    mock_tool = MagicMock()
    mock_tool._name.side_effect = lambda: "test"
    mock_base_setup_call.return_value[2] = [mock_tool]
    mock_utils.setup_call = mock_base_setup_call
    _, _, _, _, call_kwargs = setup_call(
        model="google-1.5-flash",
        client=None,
        fn=MagicMock(),
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params={},
        response_model=BaseModel,
        stream=False,
    )
    mock_client_class.assert_called_once_with()
    assert "config" in call_kwargs and isinstance(
        call_kwargs["config"], GenerateContentConfig
    )
    config = call_kwargs["config"]
    assert config.tool_config is not None
    assert config.tool_config.function_calling_config is not None
    assert config.tool_config.function_calling_config == FunctionCallingConfig(
        mode=FunctionCallingConfigMode.ANY, allowed_function_names=["test"]
    )


@patch(
    "mirascope.core.google._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.google._utils._setup_call._utils", new_callable=MagicMock)
@patch("mirascope.core.google._utils._setup_call.Client", new_callable=MagicMock)
def test_setup_call_vertexai_dynamic_config_dict(
    mock_client_class, mock_utils, mock_convert_message_params
):
    dynamic_config = {"metadata": {"tags": set()}}
    fn = MagicMock()
    base_setup = MagicMock(
        return_value=(
            "template",
            [{"role": "user", "parts": [{"text": "msg"}]}],
            None,
            {},
        )
    )
    mock_utils.setup_call = base_setup
    mock_convert_message_params.return_value = [
        {"role": "user", "parts": [{"text": "msg"}]}
    ]
    mock_client = MagicMock()
    mock_client_class.return_value = mock_client
    mock_client.vertexai = True
    _, _, _, _, _ = setup_call(  # pyright: ignore [reportCallIssue]
        model="google-1.5-flash",
        client=None,
        fn=fn,
        fn_args={},
        dynamic_config=dynamic_config,  # pyright: ignore [reportArgumentType]
        tools=None,
        json_mode=False,
        call_params={},
        response_model=None,
        stream=False,
    )
    assert "metadata" in dynamic_config
    assert "tags" in dynamic_config["metadata"]
    assert "use_vertex_ai" in dynamic_config["metadata"]["tags"]


@patch(
    "mirascope.core.google._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.google._utils._setup_call._utils", new_callable=MagicMock)
@patch("mirascope.core.google._utils._setup_call.Client", new_callable=MagicMock)
def test_setup_call_vertexai_dynamic_config_non_dict(
    mock_client_class, mock_utils, mock_convert_message_params
):
    dynamic_config = None
    fn = MagicMock()
    if hasattr(fn, "_metadata"):
        del fn._metadata
    base_setup = MagicMock(
        return_value=(
            "template",
            [{"role": "user", "parts": [{"text": "msg"}]}],
            None,
            {},
        )
    )
    mock_utils.setup_call = base_setup
    mock_convert_message_params.return_value = [
        {"role": "user", "parts": [{"text": "msg"}]}
    ]
    mock_client = MagicMock()
    mock_client_class.return_value = mock_client
    mock_client.vertexai = True
    _, _, _, _, _ = setup_call(
        model="google-1.5-flash",
        client=None,
        fn=fn,
        fn_args={},
        dynamic_config=dynamic_config,
        tools=None,
        json_mode=False,
        call_params={},
        response_model=None,
        stream=False,
    )
    assert hasattr(fn, "_metadata")
    metadata = fn._metadata
    assert "tags" in metadata
    assert "use_vertex_ai" in metadata["tags"]


@patch(
    "mirascope.core.google._utils._setup_call.convert_message_params",
    new_callable=MagicMock,
)
@patch("mirascope.core.google._utils._setup_call._utils", new_callable=MagicMock)
@patch("mirascope.core.google._utils._setup_call.Client", new_callable=MagicMock)
def test_setup_call_with_call_params(
    mock_client_class: MagicMock,
    mock_utils: MagicMock,
    mock_convert_message_params: MagicMock,
    mock_base_setup_call: MagicMock,
) -> None:
    """Tests the `setup_call` function with direct parameters in call_params."""
    mock_client = MagicMock()
    mock_client_class.return_value = mock_client
    mock_utils.setup_call = mock_base_setup_call
    fn = MagicMock()

    call_params = cast(GoogleCallParams, {"temperature": 0.8, "top_p": 0.7})

    create, prompt_template, messages, tool_types, call_kwargs = setup_call(
        model="google-1.5-flash",
        client=None,
        fn=fn,
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params=call_params,
        response_model=None,
        stream=False,
    )

    # These parameters should be passed to the base setup_call
    mock_base_setup_call.assert_called_with(
        fn,
        {},
        None,
        None,
        GoogleTool,
        call_params,  # Call params should be passed directly
        convert_common_call_params,
    )

    # Now test with a GenerateContentConfig parameter
    mock_base_setup_call.reset_mock()
    call_params = cast(GoogleCallParams, {"temperature": 0.9, "top_k": 10})

    create, prompt_template, messages, tool_types, call_kwargs = setup_call(
        model="google-1.5-flash",
        client=None,
        fn=fn,
        fn_args={},
        dynamic_config=None,
        tools=None,
        json_mode=False,
        call_params=call_params,
        response_model=None,
        stream=False,
    )

    # These parameters should be passed to the base setup_call
    mock_base_setup_call.assert_called_with(
        fn,
        {},
        None,
        None,
        GoogleTool,
        call_params,  # Call params should be passed directly
        convert_common_call_params,
    )



from typing import Literal, TypeAlias, cast

from mirascope.core.base.types import FinishReason

FinishReasonMappingValue: TypeAlias = Literal[
    "stop", "length", "tool_calls", "content_filter"
]


def _convert_finish_reasons_to_common_finish_reasons_from_mapping(
    finish_reasons: list[str] | None, mapping: dict[str, FinishReasonMappingValue]
) -> list[FinishReason] | None:
    """Provider-agnostic finish reasons."""
    if not finish_reasons:
        return None
    return [
        cast(FinishReason, mapping[reason])
        for reason in finish_reasons
        if reason in mapping
    ] or None



"""This module contains the setup_call function, which is used to set up the"""

import contextlib
from collections.abc import Awaitable, Callable, Generator
from typing import Any, cast, overload

from google.genai import Client
from google.genai.types import (
    ContentDict,
    FunctionCallingConfig,
    FunctionCallingConfigMode,
    GenerateContentConfig,
    GenerateContentConfigDict,
    GenerateContentResponse,
    Part,
    PartDict,
    ToolConfig,
    ToolListUnion,
)
from pydantic import BaseModel

from ...base import BaseMessageParam, BaseTool, _utils
from ...base._utils import (
    AsyncCreateFn,
    CreateFn,
    fn_is_async,
    get_async_create_fn,
    get_create_fn,
)
from ...base.call_params import CommonCallParams
from ...base.stream_config import StreamConfig
from .._call_kwargs import GoogleCallKwargs
from ..call_params import GoogleCallParams
from ..dynamic_config import GoogleDynamicConfig
from ..tool import GoogleTool
from ._convert_common_call_params import convert_common_call_params
from ._convert_message_params import convert_message_params


def _get_generate_content_config(
    config: GenerateContentConfig | GenerateContentConfigDict,
) -> GenerateContentConfig:
    if isinstance(config, dict):
        return GenerateContentConfig.model_validate(config)
    return config


@contextlib.contextmanager
def _generate_content_config_context(
    call_kwargs: GoogleCallKwargs,
) -> Generator[GenerateContentConfig, None, None]:
    config = call_kwargs.get("config", {})
    config = _get_generate_content_config(config)
    yield config
    call_kwargs["config"] = config


@overload
def setup_call(
    *,
    model: str,
    client: Client | None,
    fn: Callable[..., Awaitable[GoogleDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: GoogleDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: GoogleCallParams | CommonCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    AsyncCreateFn[GenerateContentResponse, GenerateContentResponse],
    str | None,
    list[ContentDict],
    list[type[GoogleTool]] | None,
    GoogleCallKwargs,
]: ...


@overload
def setup_call(
    *,
    model: str,
    client: Client | None,
    fn: Callable[..., GoogleDynamicConfig],
    fn_args: dict[str, Any],
    dynamic_config: GoogleDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: GoogleCallParams | CommonCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    CreateFn[GenerateContentResponse, GenerateContentResponse],
    str | None,
    list[ContentDict],
    list[type[GoogleTool]] | None,
    GoogleCallKwargs,
]: ...


def setup_call(
    *,
    model: str,
    client: Client | None,
    fn: Callable[..., GoogleDynamicConfig | Awaitable[GoogleDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: GoogleDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: GoogleCallParams | CommonCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    CreateFn[GenerateContentResponse, GenerateContentResponse]
    | AsyncCreateFn[GenerateContentResponse, GenerateContentResponse],
    str | None,
    list[ContentDict],
    list[type[GoogleTool]] | None,
    GoogleCallKwargs,
]:
    prompt_template, messages, tool_types, base_call_kwargs = _utils.setup_call(
        fn,
        fn_args,
        dynamic_config,
        tools,
        GoogleTool,
        call_params,
        convert_common_call_params,
    )
    call_kwargs = cast(GoogleCallKwargs, base_call_kwargs)
    messages = cast(list[BaseMessageParam | ContentDict], messages)

    if client is None:
        client = Client()

    messages = convert_message_params(messages, client)

    if messages[0] and messages[0].get("role") == "system":
        with _generate_content_config_context(call_kwargs) as config:
            system_message = messages.pop(0)
            message_parts = system_message.get("parts", [])
            if message_parts:
                config.system_instruction = [
                    Part.model_validate(part) for part in message_parts
                ]

    if json_mode:
        with _generate_content_config_context(call_kwargs) as config:
            if not tools:
                config.response_mime_type = "application/json"
        messages[-1]["parts"].append(  # pyright: ignore [reportTypedDictNotRequiredAccess, reportOptionalMemberAccess, reportArgumentType]
            PartDict(text=_utils.json_mode_content(response_model))
        )  # pyright: ignore [reportTypedDictNotRequiredAccess, reportOptionalMemberAccess, reportArgumentType]
    elif response_model:
        assert tool_types, "At least one tool must be provided for extraction."
        with _generate_content_config_context(call_kwargs) as config:
            config.tool_config = ToolConfig(
                function_calling_config=FunctionCallingConfig(
                    mode=FunctionCallingConfigMode.ANY,
                    allowed_function_names=[tool_types[0]._name()],
                )
            )
    config_tools = call_kwargs.pop("tools", None)
    if config_tools:
        with _generate_content_config_context(call_kwargs) as config:
            config.tools = cast(ToolListUnion, config_tools)

    call_kwargs |= {"model": model, "contents": messages}

    create = (
        get_async_create_fn(
            client.aio.models.generate_content,
            client.aio.models.generate_content_stream,
        )
        if fn_is_async(fn)
        else get_create_fn(
            client.models.generate_content, client.models.generate_content_stream
        )
    )
    if client.vertexai:
        if isinstance(dynamic_config, dict):
            metadata = dynamic_config.get("metadata", {})
            tags = metadata.get("tags", set())
            tags.add("use_vertex_ai")
            metadata["tags"] = tags
            dynamic_config["metadata"] = metadata
        else:
            metadata = getattr(fn, "_metadata", {})
            tags = metadata.get("tags", set())
            tags.add("use_vertex_ai")
            metadata["tags"] = tags
            fn._metadata = metadata

    return (
        create,
        prompt_template,
        messages,
        tool_types,
        call_kwargs,
    )



"""Tests the `litellm.call_response` module."""

from litellm.types.utils import ModelResponse

from mirascope.core.base.types import CostMetadata
from mirascope.core.litellm.call_response import LiteLLMCallResponse


def test_litellm_call_response_cost() -> None:
    response = ModelResponse(
        model="claude-3-5-sonnet-20240620",
        usage={"completion_tokens": 1, "prompt_tokens": 1, "total_tokens": 2},
    )
    response._hidden_params["response_cost"] = 1.8e-5
    call_response = LiteLLMCallResponse(
        metadata={},
        response=response,  # pyright: ignore [reportArgumentType]
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )

    assert call_response.cost == 1.8e-5
    assert call_response.cost_metadata == CostMetadata(cost=1.8e-5)



from .realtime import (
    Context,
    OpenAIRealtimeTool,
    Realtime,
    async_input,
    record,
    record_as_stream,
)

__all__ = [
    "Context",
    "OpenAIRealtimeTool",
    "Realtime",
    "async_input",
    "record",
    "record_as_stream",
]



"""Tests the `anthropic.call_response` module."""

from anthropic.types import (
    Message,
    MessageParam,
    TextBlock,
    ToolResultBlockParam,
    ToolUseBlock,
    Usage,
)

from mirascope.core import BaseMessageParam
from mirascope.core.anthropic.call_params import AnthropicCallParams
from mirascope.core.anthropic.call_response import AnthropicCallResponse
from mirascope.core.anthropic.tool import AnthropicTool
from mirascope.core.base import TextPart


def test_anthropic_call_response() -> None:
    """Tests the `AnthropicCallResponse` class."""
    usage = Usage(input_tokens=1, output_tokens=1)
    completion = Message(
        id="id",
        content=[TextBlock(text="content", type="text")],
        model="claude-3-5-sonnet-20240620",
        role="assistant",
        stop_reason="end_turn",
        stop_sequence=None,
        type="message",
        usage=usage,
    )
    call_response = AnthropicCallResponse(
        metadata={},
        response=completion,
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params=AnthropicCallParams(max_tokens=1000),
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    assert call_response._provider == "anthropic"
    assert call_response.content == "content"
    assert call_response.finish_reasons == ["end_turn"]
    assert call_response.model == "claude-3-5-sonnet-20240620"
    assert call_response.id == "id"
    assert call_response.usage == usage
    assert call_response.input_tokens == 1
    assert call_response.output_tokens == 1
    assert call_response.cost == 1.8e-5
    assert call_response.message_param == {
        "content": [{"text": "content", "type": "text"}],
        "role": "assistant",
    }
    assert call_response.tools is None
    assert call_response.tool is None
    assert call_response.common_finish_reasons == ["stop"]
    assert call_response.common_message_param == BaseMessageParam(
        role="assistant", content=[TextPart(type="text", text="content")]
    )
    assert call_response.common_user_message_param is None
    call_response.user_message_param = {"role": "user", "content": "content"}
    assert call_response.common_user_message_param == BaseMessageParam(
        role="user", content="content"
    )


def test_anthropic_call_response_with_tools() -> None:
    """Tests the `AnthropicCallResponse` class with tools."""

    class FormatBook(AnthropicTool):
        title: str
        author: str

        def call(self) -> str:
            return f"{self.title} by {self.author}"

    tool_call = ToolUseBlock(
        id="id",
        input={"title": "The Name of the Wind", "author": "Patrick Rothfuss"},
        name="FormatBook",
        type="tool_use",
    )
    content = TextBlock(text="content", type="text")
    completion = Message(
        id="id",
        content=[tool_call, content],
        model="claude",
        role="assistant",
        stop_reason="end_turn",
        stop_sequence=None,
        type="message",
        usage=Usage(input_tokens=1, output_tokens=2),
    )
    call_response = AnthropicCallResponse(
        metadata={},
        response=completion,
        tool_types=[FormatBook],
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params=AnthropicCallParams(max_tokens=1000),
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    tools = call_response.tools
    tool = call_response.tool
    assert tools and len(tools) == 1 and tools[0] == tool
    assert isinstance(tool, FormatBook)
    assert tool.title == "The Name of the Wind"
    assert tool.author == "Patrick Rothfuss"
    output = tool.call()
    assert output == "The Name of the Wind by Patrick Rothfuss"
    assert call_response.tool_message_params([(tool, output)]) == [
        MessageParam(
            role="user",
            content=[
                ToolResultBlockParam(
                    tool_use_id="id",
                    type="tool_result",
                    content=[{"text": output, "type": "text"}],
                )
            ],
        )
    ]



"""Calculate the cost of a completion using the Vertex AI Gemini API, considering context window size."""

from ..base.types import CostMetadata


def calculate_cost(
    metadata: CostMetadata,
    model: str = "gemini-1.5-pro",
) -> float | None:
    """Calculate the cost of a completion using the Vertex AI Gemini API.

    https://cloud.google.com/vertex-ai/pricing#generative_ai_models

    Model               Input (<=128K)     Output (<=128K)    Input (>128K)      Output (>128K)
    gemini-1.5-flash    $0.00001875 / 1k   $0.000075 / 1k     $0.0000375 / 1k    $0.00015 / 1k
    gemini-1.5-pro      $0.00125 / 1k      $0.00375 / 1k      $0.0025 / 1k       $0.0075 / 1k
    gemini-1.0-pro      $0.000125 / 1k     $0.000375 / 1k     N/A                N/A

    Note: Prices are per 1k characters. Gemini 1.0 Pro only supports up to 32K context window.
    """

    context_length = metadata.context_length or 0
    pricing = {
        "gemini-1.5-flash": {
            "prompt_short": 0.000_018_75,
            "completion_short": 0.000_075,
            "prompt_long": 0.000_037_5,
            "completion_long": 0.000_15,
        },
        "gemini-1.5-pro": {
            "prompt_short": 0.001_25,
            "completion_short": 0.003_75,
            "prompt_long": 0.002_5,
            "completion_long": 0.007_5,
        },
        "gemini-1.0-pro": {
            "prompt_short": 0.000_125,
            "completion_short": 0.000_375,
            "prompt_long": None,
            "completion_long": None,
        },
    }

    if metadata.input_tokens is None or metadata.output_tokens is None:
        return None

    try:
        model_pricing = pricing[model]
    except KeyError:
        return None

    # Determine if we're using long context pricing
    use_long_context = context_length > 128000

    if use_long_context and model == "gemini-1.0-pro":
        return None  # Gemini 1.0 Pro doesn't support long context

    prompt_price = model_pricing["prompt_long" if use_long_context else "prompt_short"]
    completion_price = model_pricing[
        "completion_long" if use_long_context else "completion_short"
    ]

    prompt_cost = (metadata.input_tokens / 1000) * prompt_price
    completion_cost = (metadata.output_tokens / 1000) * completion_price
    total_cost = prompt_cost + completion_cost

    return total_cost



from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Literal, TypeAlias

from pydantic import BaseModel, ConfigDict, Field

if TYPE_CHECKING:
    from PIL import Image
    from pydub import AudioSegment

    has_pil_module: bool
    has_pydub_module: bool
else:
    try:
        from PIL import Image  # pyright: ignore [reportAssignmentType]

        has_pil_module = True
    except ImportError:  # pragma: no cover
        has_pil_module = False

        class Image:
            class Image:
                def tobytes(self) -> bytes: ...

    try:
        from pydub import AudioSegment  # pyright: ignore [reportAssignmentType]

        has_pydub_module = True
    except ImportError:  # pragma: no cover
        has_pydub_module = False

        from io import FileIO

        class AudioSegment:
            def set_frame_rate(self, rate: int) -> AudioSegment: ...
            def set_channels(self, channels: int) -> AudioSegment: ...
            def set_sample_width(self, sample_width: int) -> AudioSegment: ...
            def export(self, format: str) -> FileIO: ...
            def read(self) -> bytes: ...


FinishReason: TypeAlias = Literal["stop", "length", "tool_calls", "content_filter"]


class Usage(BaseModel):
    input_tokens: int
    """Number of tokens in the prompt."""

    cached_tokens: int
    """Number of tokens used that were previously cached (and thus cheaper)."""

    output_tokens: int
    """Number of tokens in the generated output."""

    total_tokens: int
    """Total number of tokens used in the request (prompt + completion)."""


JsonableType: TypeAlias = (
    str
    | int
    | float
    | bool
    | bytes
    | list["JsonableType"]
    | set["JsonableType"]
    | tuple["JsonableType", ...]
    | dict[str, "JsonableType"]
    | BaseModel
)


class VideoMetadata(BaseModel):
    """Metadata for a video for cost calculation"""

    duration_seconds: Annotated[
        float,
        Field(description="Duration of the video in seconds"),
    ]

    with_audio: Annotated[
        bool | None,
        Field(
            default=False,
            description="Whether the video includes audio that should be processed",
        ),
    ] = False

    tokens: Annotated[
        int | None,
        Field(default=None, description="Precalculated token count for this video"),
    ] = None


class AudioMetadata(BaseModel):
    """Metadata for an audio file for cost calculation"""

    duration_seconds: Annotated[
        float,
        Field(description="Duration of the audio in seconds"),
    ]

    with_timestamps: Annotated[
        bool | None,
        Field(default=False, description="Whether timestamps should be included"),
    ] = False

    tokens: Annotated[
        int | None,
        Field(default=None, description="Precalculated token count for this audio"),
    ] = None


class ImageMetadata(BaseModel):
    """Metadata for an image for cost calculation"""

    width: Annotated[
        int,
        Field(description="Width of the image in pixels"),
    ]

    height: Annotated[
        int,
        Field(description="Height of the image in pixels"),
    ]

    tokens: Annotated[
        int | None,
        Field(default=None, description="Precalculated token count for this image"),
    ] = None

    detail: Annotated[
        str | None,
        Field(default=None, description="Detail level of the image"),
    ] = None


class GoogleMetadata(BaseModel):
    """Google API specific metadata for cost calculation"""

    use_vertex_ai: Annotated[
        bool | None,
        Field(
            default=False,
            description="Whether to use Vertex AI pricing (vs. direct Gemini API)",
        ),
    ] = False

    grounding_requests: Annotated[
        int | None,
        Field(default=None, description="Number of Google Search grounding requests"),
    ] = None


class PDFImageMetadata(BaseModel):
    """Metadata for an image extracted from a PDF page"""

    width: Annotated[
        int,
        Field(description="Width of the image in pixels"),
    ]

    height: Annotated[
        int,
        Field(description="Height of the image in pixels"),
    ]

    tokens: Annotated[
        int | None,
        Field(default=None, description="Precalculated token count for this image"),
    ] = None


class PDFMetadata(BaseModel):
    """Metadata specific to PDF documents for cost calculation"""

    page_count: Annotated[
        int | None,
        Field(default=None, description="Number of pages in the PDF"),
    ] = None

    text_tokens: Annotated[
        int | None,
        Field(
            default=None, description="Number of tokens from text content in the PDF"
        ),
    ] = None

    images: Annotated[
        list[PDFImageMetadata] | None,
        Field(
            default=None,
            description="List of images extracted from PDF with width and height information",
        ),
    ] = None

    cached: Annotated[
        bool | None,
        Field(
            default=None,
            description="Whether this PDF was cached for reduced token costs",
        ),
    ] = None


class CostMetadata(BaseModel):
    """Metadata required for accurate LLM API cost calculation across all providers."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    # Common fields
    input_tokens: Annotated[
        int | float | None,
        Field(default=None, description="Input tokens"),
    ] = None
    output_tokens: Annotated[
        int | float | None,
        Field(default=None, description="Output tokens"),
    ] = None
    cached_tokens: Annotated[
        int | float | None,
        Field(default=None, description="Cached tokens"),
    ] = None
    streaming_mode: Annotated[
        bool | None,
        Field(default=None, description="Whether streaming API was used"),
    ] = None
    cached_response: Annotated[
        bool | None,
        Field(default=None, description="Whether response was served from cache"),
    ] = None
    context_length: Annotated[
        int | None,
        Field(default=None, description="Total context window length in tokens"),
    ] = None
    realtime_mode: Annotated[
        bool | None,
        Field(default=None, description="Whether realtime processing was used"),
    ] = None
    region: Annotated[
        str | None,
        Field(
            default=None,
            description="Cloud region for request (affects pricing in some providers)",
        ),
    ] = None
    tier: Annotated[
        str | None,
        Field(default=None, description="Service tier (e.g. standard, enterprise)"),
    ] = None
    batch_mode: Annotated[
        bool | None,
        Field(
            default=False,
            description="Whether batch mode is used (discount usually applies)",
        ),
    ] = None

    # Media-related fields
    images: Annotated[
        list[ImageMetadata] | None,
        Field(default=None, description="List of images with their metadata"),
    ] = None
    videos: Annotated[
        list[VideoMetadata] | None,
        Field(default=None, description="List of videos with their metadata"),
    ] = None
    audio: Annotated[
        list[AudioMetadata] | None,
        Field(default=None, description="List of audio clips with their metadata"),
    ] = None
    audio_output: Annotated[
        list[AudioMetadata] | None,
        Field(
            default=None, description="List of audio output clips with their metadata"
        ),
    ] = None
    # PDF-related fields
    pdf: Annotated[
        PDFMetadata | None,
        Field(default=None, description="Metadata for PDF documents"),
    ] = None

    # Context caching related fields
    context_cache_tokens: Annotated[
        int | None,
        Field(default=None, description="Number of cached context tokens"),
    ] = None
    context_cache_hours: Annotated[
        float | None,
        Field(default=None, description="Number of hours to keep context in cache"),
    ] = None

    # Provider-specific fields
    google: Annotated[
        GoogleMetadata | None,
        Field(
            default=None,
            description="Google/Gemini-specific metadata for cost calculation",
        ),
    ] = None
    realtime_tokens: Annotated[
        int | None,
        Field(
            default=None,
            description="[OpenAI] Number of realtime tokens in the request",
        ),
    ] = None

    # Anthropic-specific fields
    cache_write: Annotated[
        bool | None,
        Field(default=None, description="[Anthropic] Whether cache write occurred"),
    ] = None
    tool_use_tokens: Annotated[
        int | None,
        Field(default=None, description="[Anthropic] Tokens used for tool calls"),
    ] = None

    # If the provider happens to provide the cost, we should just use that.
    cost: Annotated[
        float | None,
        Field(default=None, description="Cost provided by the API response"),
    ] = None


Provider: TypeAlias = Literal[
    "anthropic",
    "azure",
    "bedrock",
    "cohere",
    "gemini",
    "google",
    "groq",
    "litellm",
    "mistral",
    "openai",
    "vertex",
    "xai",
]
LocalProvider: TypeAlias = Literal[
    "ollama",
    "vllm",
]



"""This module contains the `AnthropicCallResponse` class.

usage docs: learn/calls.md#handling-responses
"""

from collections.abc import Sequence
from functools import cached_property

from anthropic.types import (
    Message,
    MessageParam,
    ToolParam,
    ToolResultBlockParam,
    Usage,
)
from pydantic import SerializeAsAny, computed_field

from .. import BaseMessageParam
from ..base import BaseCallResponse, transform_tool_outputs, types
from ..base.types import CostMetadata
from ._utils._convert_finish_reason_to_common_finish_reasons import (
    _convert_finish_reasons_to_common_finish_reasons,
)
from ._utils._message_param_converter import AnthropicMessageParamConverter
from .call_params import AnthropicCallParams
from .dynamic_config import AnthropicDynamicConfig, AsyncAnthropicDynamicConfig
from .tool import AnthropicTool


class AnthropicCallResponse(
    BaseCallResponse[
        Message,
        AnthropicTool,
        ToolParam,
        AsyncAnthropicDynamicConfig | AnthropicDynamicConfig,
        MessageParam,
        AnthropicCallParams,
        MessageParam,
        AnthropicMessageParamConverter,
    ]
):
    """A convenience wrapper around the Anthropic `Message` response.

    When calling the Anthropic API using a function decorated with `anthropic_call`, the
    response will be an `AnthropicCallResponse` instance with properties that allow for
    more convenience access to commonly used attributes.

    Example:

    ```python
    from mirascope.core import prompt_template
    from mirascope.core.anthropic import anthropic_call


    @anthropic_call("claude-3-5-sonnet-20240620")
    def recommend_book(genre: str) -> str:
        return f"Recommend a {genre} book"


    response = recommend_book("fantasy")  # response is an `AnthropicCallResponse` instance
    print(response.content)
    ```
    """

    _message_converter: type[AnthropicMessageParamConverter] = (
        AnthropicMessageParamConverter
    )

    _provider = "anthropic"

    @computed_field
    @property
    def content(self) -> str:
        """Returns the string text of the 0th text block."""
        block = self.response.content[0]
        return block.text if block.type == "text" else ""

    @computed_field
    @property
    def finish_reasons(self) -> list[str]:
        """Returns the finish reasons of the response."""
        return [str(self.response.stop_reason)]

    @computed_field
    @property
    def model(self) -> str:
        """Returns the name of the response model."""
        return self.response.model

    @computed_field
    @property
    def id(self) -> str:
        """Returns the id of the response."""
        return self.response.id

    @property
    def usage(self) -> Usage:
        """Returns the usage of the message."""
        return self.response.usage

    @computed_field
    @property
    def input_tokens(self) -> int:
        """Returns the number of input tokens."""
        return self.usage.input_tokens

    @computed_field
    @property
    def cached_tokens(self) -> int:
        """Returns the number of cached tokens."""
        return getattr(self.usage, "cache_read_input_tokens", 0)

    @computed_field
    @property
    def output_tokens(self) -> int:
        """Returns the number of output tokens."""
        return self.usage.output_tokens

    @computed_field
    @cached_property
    def message_param(self) -> SerializeAsAny[MessageParam]:
        """Returns the assistants's response as a message parameter."""
        return MessageParam(**self.response.model_dump(include={"content", "role"}))

    @cached_property
    def tools(self) -> list[AnthropicTool] | None:
        """Returns any available tool calls as their `AnthropicTool` definition.

        Raises:
            ValidationError: if a tool call doesn't match the tool's schema.
        """
        if not self.tool_types:
            return None

        extracted_tools = []
        for content in self.response.content:
            if content.type != "tool_use":
                continue
            for tool_type in self.tool_types:
                if content.name == tool_type._name():
                    extracted_tools.append(tool_type.from_tool_call(content))
                    break

        return extracted_tools

    @cached_property
    def tool(self) -> AnthropicTool | None:
        """Returns the 0th tool for the 0th choice message.

        Raises:
            ValidationError: if the tool call doesn't match the tool's schema.
        """
        if tools := self.tools:
            return tools[0]
        return None

    @classmethod
    @transform_tool_outputs
    def tool_message_params(
        cls, tools_and_outputs: Sequence[tuple[AnthropicTool, str]]
    ) -> list[MessageParam]:
        """Returns the tool message parameters for tool call results.

        Args:
            tools_and_outputs: The sequence of tools and their outputs from which the tool
                message parameters should be constructed.

        Returns:
            The list of constructed `MessageParam` parameters.
        """
        return [
            {
                "role": "user",
                "content": [
                    ToolResultBlockParam(
                        tool_use_id=tool.tool_call.id,  # pyright: ignore [reportOptionalMemberAccess]
                        type="tool_result",
                        content=[{"text": output, "type": "text"}],
                    )
                    for tool, output in tools_and_outputs
                ],
            }
        ]

    @property
    def common_finish_reasons(self) -> list[types.FinishReason] | None:
        """Provider-agnostic finish reasons."""
        return _convert_finish_reasons_to_common_finish_reasons(self.finish_reasons)

    @property
    def common_message_param(self) -> BaseMessageParam:
        return AnthropicMessageParamConverter.from_provider([(self.message_param)])[0]

    @property
    def common_user_message_param(self) -> BaseMessageParam | None:
        if not self.user_message_param:
            return None
        return AnthropicMessageParamConverter.from_provider(
            [(self.user_message_param)]
        )[0]

    @computed_field
    @property
    def cost_metadata(self) -> CostMetadata:
        """Get metadata required for cost calculation."""
        return super().cost_metadata



"""Tests the `google._utils.get_json_output` module."""

import pytest
from google.genai.types import (
    Candidate,
    Content,
    FunctionCall,
    GenerateContentResponse,
    Part,
)
from google.genai.types import (
    FinishReason as GoogleFinishReason,
)

from mirascope.core.google._utils._get_json_output import get_json_output
from mirascope.core.google.call_response import GoogleCallResponse
from mirascope.core.google.call_response_chunk import GoogleCallResponseChunk


@pytest.fixture()
def mock_generate_content_response() -> GenerateContentResponse:
    """Returns a mock `GenerateContentResponse` instance."""
    return GenerateContentResponse(
        candidates=[
            Candidate(
                finish_reason=GoogleFinishReason.STOP,
                content=Content(
                    parts=[
                        Part(text='{"key": "value"}'),
                        Part(
                            function_call=FunctionCall(
                                name="FormatBook",
                                args={
                                    "title": "The Name of the Wind",
                                    "author": "Patrick Rothfuss",
                                },
                            )
                        ),
                    ],
                    role="model",
                ),
            )
        ]
    )


def test_get_json_output_call_response(
    mock_generate_content_response: GenerateContentResponse,
) -> None:
    """Tests the `get_json_output` function with a call response."""
    call_response = GoogleCallResponse(
        metadata={},
        response=mock_generate_content_response,
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    assert get_json_output(call_response, json_mode=True) == '{"key": "value"}'
    assert (
        get_json_output(call_response, json_mode=False)
        == '{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}'
    )

    mock_generate_content_response.candidates[0].content.parts.pop(1)  # pyright: ignore [reportOptionalMemberAccess, reportOptionalSubscript]
    with pytest.raises(
        ValueError, match="No tool call or JSON object found in response."
    ):
        get_json_output(call_response, json_mode=False)


def test_get_json_output_call_response_chunk(
    mock_generate_content_response: GenerateContentResponse,
) -> None:
    """Tests the `get_json_output` function with a call response chunk."""
    call_response_chunk = GoogleCallResponseChunk(chunk=mock_generate_content_response)
    assert get_json_output(call_response_chunk, json_mode=True) == '{"key": "value"}'

    with pytest.raises(
        ValueError, match="Google only supports structured streaming in json mode."
    ):
        get_json_output(call_response_chunk, json_mode=False)



"""Tool class for the provider-agnostic tool."""

from __future__ import annotations

from typing import Any, TypeVar

from pydantic._internal._model_construction import ModelMetaclass

from ..core.base.tool import BaseTool

_ToolResponseT = TypeVar("_ToolResponseT", bound=BaseTool)


class _DelegateAbstractMethodsForTool(ModelMetaclass):
    def __new__(
        mcls,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        **kwargs: Any,  # noqa: ANN401
    ) -> type:
        cls = super().__new__(mcls, name, bases, namespace)
        cls.__abstractmethods__ = frozenset()
        return cls


_ToolMessageParamT = TypeVar("_ToolMessageParamT")


class Tool(BaseTool, metaclass=_DelegateAbstractMethodsForTool):
    """
    A provider-agnostic Tool class.
    - No BaseProviderConverter
    - Relies on _response having `common_` methods/properties if needed.
    """

    _tool: BaseTool

    def __init__(self, tool: BaseTool) -> None:
        super().__init__(**{field: getattr(tool, field) for field in tool.model_fields})
        object.__setattr__(self, "_tool", tool)

    def __getattribute__(self, name: str) -> Any:  # noqa: ANN401
        special_names = {
            "_response",
            "__dict__",
            "__class__",
            "model_fields",
            "__annotations__",
            "__pydantic_validator__",
            "__pydantic_fields_set__",
            "__pydantic_extra__",
            "__pydantic_private__",
            "__class_getitem__",
        }

        if name in special_names:
            return object.__getattribute__(self, name)

        try:
            tool = object.__getattribute__(self, "_tool")
            return getattr(tool, name)
        except AttributeError:
            return object.__getattribute__(self, name)



"""This module defines the function return type for functions as LLM calls."""

from cohere import (
    AsyncClient,
    Client,
)
from cohere.types.chat_message import ChatMessage

from ..base import BaseDynamicConfig, BaseMessageParam
from .call_params import CohereCallParams

AsyncCohereDynamicConfig = BaseDynamicConfig[
    ChatMessage | BaseMessageParam, CohereCallParams, AsyncClient
]
CohereDynamicConfig = BaseDynamicConfig[
    ChatMessage | BaseMessageParam, CohereCallParams, Client
]
"""The function return type for functions wrapped with the `cohere_call` decorator.

Example:

```python
from mirascope.core import prompt_template
from mirascope.core.cohere import CohereDynamicConfig, cohere_call


@cohere_call("command-r-plus")
@prompt_template("Recommend a {capitalized_genre} book")
def recommend_book(genre: str) -> CohereDynamicConfig:
    return {"computed_fields": {"capitalized_genre": genre.capitalize()}}
```
"""



"""Tests the `_google_calculate_cost` function."""

from math import isclose

from mirascope.core.base.types import CostMetadata, GoogleMetadata
from mirascope.core.costs._google_calculate_cost import (
    GEMINI_API_PRICING,
    VERTEX_AI_PRICING,
    _calculate_context_cache_cost,
    _calculate_grounding_cost,
    _calculate_standard_gemini_cost,
    _calculate_vertex_1_5_cost,
    _calculate_vertex_2_0_cost,
    calculate_cost,
)


def test_calculate_cost() -> None:
    """Tests the `calculate_cost` function."""
    # Test invalid inputs
    assert calculate_cost(CostMetadata(), model="google-1.5-flash") is None
    assert (
        calculate_cost(CostMetadata(input_tokens=1, output_tokens=1), model="unknown")
        is None
    )

    # Test basic Gemini 1.5 Flash model
    assert (
        calculate_cost(
            CostMetadata(input_tokens=1, output_tokens=1), model="gemini-1.5-flash"
        )
        == 0.000000375
    )

    # Test short context pricing
    assert (
        calculate_cost(
            CostMetadata(input_tokens=1000, output_tokens=500), model="gemini-1.5-pro"
        )
        == 1000 * 0.00000125 + 500 * 0.000005
    )

    # Test long context pricing (input > 128K tokens)
    assert (
        calculate_cost(
            CostMetadata(input_tokens=130000, output_tokens=1000),
            model="gemini-1.5-pro",
        )
        == 130000 * 0.0000025 + 1000 * 0.00001
    )

    # Test with cached tokens
    assert (
        calculate_cost(
            CostMetadata(input_tokens=1000, output_tokens=500, cached_tokens=200),
            model="gemini-1.5-pro",
        )
        == 1000 * 0.00000125 + 500 * 0.000005 + 200 * 0.000000625
    )

    # Test context cache costs
    assert (
        calculate_cost(
            CostMetadata(
                input_tokens=1000,
                output_tokens=500,
                context_cache_tokens=10000,
                context_cache_hours=2,
            ),
            model="gemini-1.5-pro",
        )
        == 1000 * 0.00000125 + 500 * 0.000005 + 10000 * 0.000001 * 2
    )

    # Test Vertex AI pricing for Gemini 2.0
    assert (
        calculate_cost(
            CostMetadata(
                input_tokens=1000,
                output_tokens=500,
                google=GoogleMetadata(use_vertex_ai=True),
            ),
            model="gemini-2.0-flash",
        )
        == 1000 * 0.00000015 + 500 * 0.0000006
    )

    # Test batch mode discount (50% for Vertex AI)
    assert (
        calculate_cost(
            CostMetadata(
                input_tokens=1000,
                output_tokens=500,
                batch_mode=True,
                google=GoogleMetadata(use_vertex_ai=True),
            ),
            model="gemini-2.0-flash",
        )
        == (1000 * 0.00000015 + 500 * 0.0000006) * 0.5
    )

    # Test with grounding requests (for Gemini 2.0 with Vertex AI)
    assert (
        calculate_cost(
            CostMetadata(
                input_tokens=1000,
                output_tokens=500,
                google=GoogleMetadata(use_vertex_ai=True, grounding_requests=2000),
            ),
            model="gemini-2.0-flash",
        )
        == 1000 * 0.00000015 + 500 * 0.0000006 + (500 / 1000) * 35.0
    )

    # Test free tier for grounding requests (first 1500 requests)
    assert (
        calculate_cost(
            CostMetadata(
                input_tokens=1000,
                output_tokens=500,
                google=GoogleMetadata(use_vertex_ai=True, grounding_requests=1000),
            ),
            model="gemini-2.0-flash",
        )
        == 1000 * 0.00000015 + 500 * 0.0000006
    )

    # Test Gemini 1.0 model pricing
    assert (
        calculate_cost(
            CostMetadata(input_tokens=1000, output_tokens=500), model="gemini-1.0-pro"
        )
        == 1000 * 0.0000005 + 500 * 0.0000015
    )


def test_context_cache_cost_none_tokens():
    meta = CostMetadata(
        input_tokens=100,
        output_tokens=100,
        context_cache_tokens=None,
        context_cache_hours=2,
    )
    pricing = GEMINI_API_PRICING["gemini-1.5-pro"]
    cost = _calculate_context_cache_cost(
        meta, pricing, "gemini-1.5-pro", use_vertex_ai=False
    )
    assert cost == 0.0


def test_context_cache_cost_none_hours():
    meta = CostMetadata(
        input_tokens=100,
        output_tokens=100,
        context_cache_tokens=10000,
        context_cache_hours=None,
    )
    pricing = GEMINI_API_PRICING["gemini-1.5-pro"]
    cost = _calculate_context_cache_cost(
        meta, pricing, "gemini-1.5-pro", use_vertex_ai=False
    )
    assert cost == 0.0


def test_context_cache_cost_vertex_2_0():
    meta = CostMetadata(
        input_tokens=100,
        output_tokens=100,
        context_cache_tokens=10000,
        context_cache_hours=2,
    )
    pricing = VERTEX_AI_PRICING[
        "gemini-2.0-flash"
    ]  # cache_storage_per_hour = 0.00000100
    cost = _calculate_context_cache_cost(
        meta, pricing, "gemini-2.0-flash", use_vertex_ai=True
    )
    expected = (
        10000 * pricing.get("cache_storage_per_hour", 0) * 2
    )  # 10000*0.00000100*2 = 0.02
    assert isclose(cost, expected, rel_tol=1e-9)


def test_context_cache_cost_vertex_1_5():
    meta = CostMetadata(
        input_tokens=100,
        output_tokens=100,
        context_cache_tokens=10000,
        context_cache_hours=2,
    )
    pricing = VERTEX_AI_PRICING["gemini-1.5-pro"]  # cache_storage_text = 0.001125
    cost = _calculate_context_cache_cost(
        meta, pricing, "gemini-1.5-pro", use_vertex_ai=True
    )
    expected = 40 * pricing["cache_storage_text"] * 2
    assert isclose(cost, expected, rel_tol=1e-9)


def test_context_cache_cost_standard():
    meta = CostMetadata(
        input_tokens=100,
        output_tokens=100,
        context_cache_tokens=10000,
        context_cache_hours=2,
    )
    pricing = GEMINI_API_PRICING["gemini-1.5-pro"]
    # Standard branch: storage_rate_per_token = 0.000001
    expected = 10000 * 0.000001 * 2  # = 0.02
    cost = _calculate_context_cache_cost(
        meta, pricing, "gemini-1.5-pro", use_vertex_ai=False
    )
    assert isclose(cost, expected, rel_tol=1e-9)


def test_context_cache_cost_standard_flash_8b():
    meta = CostMetadata(
        input_tokens=100,
        output_tokens=100,
        context_cache_tokens=10000,
        context_cache_hours=2,
    )
    pricing = GEMINI_API_PRICING["gemini-1.5-flash-8b"]
    expected = 10000 * 0.00000025 * 2  # = 0.005
    cost = _calculate_context_cache_cost(
        meta, pricing, "gemini-1.5-flash-8b", use_vertex_ai=False
    )
    assert isclose(cost, expected, rel_tol=1e-9)


def test_grounding_cost_no_google():
    meta = CostMetadata(input_tokens=100, output_tokens=100)
    cost = _calculate_grounding_cost(meta, "gemini-2.0-flash")
    assert cost == 0.0


def test_grounding_cost_no_requests():
    meta = CostMetadata(
        input_tokens=100,
        output_tokens=100,
        google=GoogleMetadata(use_vertex_ai=True, grounding_requests=None),
    )
    cost = _calculate_grounding_cost(meta, "gemini-2.0-flash")
    assert cost == 0.0


def test_grounding_cost_free_tier():
    meta = CostMetadata(
        input_tokens=100,
        output_tokens=100,
        google=GoogleMetadata(use_vertex_ai=True, grounding_requests=1500),
    )
    cost = _calculate_grounding_cost(meta, "gemini-2.0-flash")
    assert cost == 0.0


def test_grounding_cost_excess_vertex():
    meta = CostMetadata(
        input_tokens=100,
        output_tokens=100,
        google=GoogleMetadata(use_vertex_ai=True, grounding_requests=2000),
    )
    # Excess = 2000 - 1500 = 500; cost = (500/1000)*35 = 17.5
    cost = _calculate_grounding_cost(meta, "gemini-2.0-flash")
    assert isclose(cost, 17.5, rel_tol=1e-9)


def test_grounding_cost_non_vertex():
    meta = CostMetadata(
        input_tokens=100,
        output_tokens=100,
        google=GoogleMetadata(use_vertex_ai=False, grounding_requests=2000),
    )
    # For non–Vertex AI, no free tier: cost = (2000/1000)*35 = 70
    cost = _calculate_grounding_cost(meta, "gemini-2.0-flash")
    assert isclose(cost, 70.0, rel_tol=1e-9)


def test_grounding_cost_other_model():
    meta = CostMetadata(
        input_tokens=100,
        output_tokens=100,
        google=GoogleMetadata(use_vertex_ai=True, grounding_requests=2000),
    )
    # For a model other than gemini-2.0-flash, excess_requests = grounding_requests = 2000
    cost = _calculate_grounding_cost(meta, "gemini-1.5-flash")
    expected = (2000 / 1000) * 35.0  # = 70.0
    assert isclose(cost, expected, rel_tol=1e-9)


def test_vertex_2_0_cost_non_batch():
    meta = CostMetadata(
        input_tokens=1000,
        output_tokens=500,
        cached_tokens=100,
        context_cache_tokens=None,
        context_cache_hours=None,
        google=GoogleMetadata(use_vertex_ai=True, grounding_requests=None),
        batch_mode=False,
    )
    pricing = VERTEX_AI_PRICING["gemini-2.0-flash"]
    cost = _calculate_vertex_2_0_cost(meta, pricing, "gemini-2.0-flash")
    expected = (
        1000 * pricing["text_input"] + 500 * pricing["output"] + 100 * pricing["cached"]
    )
    # No context cache (tokens None) and grounding cost is 0
    assert isclose(cost, expected, rel_tol=1e-9)


def test_vertex_2_0_cost_batch():
    meta = CostMetadata(
        input_tokens=1000,
        output_tokens=500,
        cached_tokens=100,
        context_cache_tokens=10000,
        context_cache_hours=2,
        google=GoogleMetadata(use_vertex_ai=True, grounding_requests=2000),
        batch_mode=True,
    )
    pricing = VERTEX_AI_PRICING["gemini-2.0-flash"]
    # Compute individual components
    prompt_cost = 1000 * pricing["text_input"]
    completion_cost = 500 * pricing["output"]
    cached_cost = 100 * pricing["cached"]
    context_cache_cost = _calculate_context_cache_cost(
        meta, pricing, "gemini-2.0-flash", True
    )
    grounding_cost = _calculate_grounding_cost(meta, "gemini-2.0-flash")
    # Batch discount applies to prompt, completion, and context cache
    prompt_cost *= 0.5
    completion_cost *= 0.5
    context_cache_cost *= 0.5
    expected = (
        prompt_cost
        + completion_cost
        + cached_cost
        + context_cache_cost
        + grounding_cost
    )
    cost = _calculate_vertex_2_0_cost(meta, pricing, "gemini-2.0-flash")
    assert isclose(cost, expected, rel_tol=1e-9)


def test_vertex_1_5_cost_non_batch():
    meta = CostMetadata(
        input_tokens=1000,
        output_tokens=500,
        cached_tokens=0,
        context_cache_tokens=None,
        context_cache_hours=None,
        google=GoogleMetadata(use_vertex_ai=True, grounding_requests=None),
        batch_mode=False,
    )
    pricing = VERTEX_AI_PRICING["gemini-1.5-pro"]
    # text_cost = (1000*4/1000)*pricing["text_input"]
    # output_cost = (500*4/1000)*pricing["output"]
    expected = (1000 * 4 / 1000) * pricing["text_input"] + (500 * 4 / 1000) * pricing[
        "output"
    ]
    cost = _calculate_vertex_1_5_cost(meta, pricing, "gemini-1.5-pro")
    assert isclose(cost, expected, rel_tol=1e-9)


def test_vertex_1_5_cost_batch():
    meta = CostMetadata(
        input_tokens=1000,
        output_tokens=500,
        cached_tokens=0,
        context_cache_tokens=10000,
        context_cache_hours=2,
        google=GoogleMetadata(use_vertex_ai=True, grounding_requests=2000),
        batch_mode=True,
    )
    pricing = VERTEX_AI_PRICING["gemini-1.5-pro"]
    text_cost = (1000 * 4 / 1000) * pricing["text_input"]
    output_cost = (500 * 4 / 1000) * pricing["output"]
    context_cache_cost = _calculate_context_cache_cost(
        meta, pricing, "gemini-1.5-pro", True
    )
    grounding_cost = _calculate_grounding_cost(meta, "gemini-1.5-pro")
    text_cost *= 0.5
    output_cost *= 0.5
    context_cache_cost *= 0.5
    expected = text_cost + output_cost + context_cache_cost + grounding_cost
    cost = _calculate_vertex_1_5_cost(meta, pricing, "gemini-1.5-pro")
    assert isclose(cost, expected, rel_tol=1e-9)


def test_standard_gemini_cost_short_context():
    meta = CostMetadata(
        input_tokens=1000,
        output_tokens=500,
        cached_tokens=200,
        context_cache_tokens=10000,
        context_cache_hours=2,
        google=None,
        batch_mode=False,
    )
    pricing = GEMINI_API_PRICING["gemini-1.5-pro"]
    use_long_context = False  # input_tokens is low and context_length is not set
    prompt_price = pricing["prompt_short"]
    cached_price = pricing["cached"]
    completion_price = pricing["completion_short"]
    prompt_cost = 1000 * prompt_price
    cached_cost = 200 * cached_price
    completion_cost = 500 * completion_price
    context_cache_cost = _calculate_context_cache_cost(
        meta, pricing, "gemini-1.5-pro", use_vertex_ai=False
    )
    grounding_cost = _calculate_grounding_cost(meta, "gemini-1.5-pro")
    expected = (
        prompt_cost
        + cached_cost
        + completion_cost
        + context_cache_cost
        + grounding_cost
    )
    cost = _calculate_standard_gemini_cost(
        meta, pricing, "gemini-1.5-pro", use_long_context
    )
    assert isclose(cost, expected, rel_tol=1e-9)


def test_standard_gemini_cost_long_context():
    meta = CostMetadata(
        input_tokens=1000,
        output_tokens=500,
        cached_tokens=200,
        context_cache_tokens=10000,
        context_cache_hours=2,
        context_length=130000,  # triggers long context pricing
        google=None,
        batch_mode=False,
    )
    pricing = GEMINI_API_PRICING["gemini-1.5-pro"]
    use_long_context = True
    prompt_price = pricing["prompt_long"]
    cached_price = pricing["cached"]
    completion_price = pricing["completion_long"]
    prompt_cost = 1000 * prompt_price
    cached_cost = 200 * cached_price
    completion_cost = 500 * completion_price
    context_cache_cost = _calculate_context_cache_cost(
        meta, pricing, "gemini-1.5-pro", use_vertex_ai=False
    )
    grounding_cost = _calculate_grounding_cost(meta, "gemini-1.5-pro")
    expected = (
        prompt_cost
        + cached_cost
        + completion_cost
        + context_cache_cost
        + grounding_cost
    )
    cost = _calculate_standard_gemini_cost(
        meta, pricing, "gemini-1.5-pro", use_long_context
    )
    assert isclose(cost, expected, rel_tol=1e-9)


def test_calculate_cost_vertex_1_5():
    meta = CostMetadata(
        input_tokens=1000,
        output_tokens=500,
        google=GoogleMetadata(use_vertex_ai=True, grounding_requests=0),
        batch_mode=False,
    )
    pricing = VERTEX_AI_PRICING["gemini-1.5-pro"]
    expected = _calculate_vertex_1_5_cost(meta, pricing, "gemini-1.5-pro")
    cost = calculate_cost(meta, "gemini-1.5-pro")
    assert isclose(cost, expected, rel_tol=1e-9)  # pyright: ignore [reportArgumentType]


def test_calculate_cost_vertex_2_0() -> None:
    """Test that calculate_cost uses the Vertex AI branch for Gemini 2.0 models."""
    meta = CostMetadata(
        input_tokens=1000,
        output_tokens=500,
        cached_tokens=100,
        context_cache_tokens=2000,
        context_cache_hours=1,
        google=GoogleMetadata(use_vertex_ai=True, grounding_requests=0),
        batch_mode=False,
    )
    model = "gemini-2.0-flash"  # a model name starting with "gemini-2.0"
    pricing = VERTEX_AI_PRICING[model]
    expected = _calculate_vertex_2_0_cost(meta, pricing, model)
    cost = calculate_cost(meta, model)
    assert isclose(cost, expected, rel_tol=1e-9)  # pyright: ignore [reportArgumentType]


def test_calculate_cost_vertex_unsupported_model() -> None:
    """Test that calculate_cost correctly dispatches to None for unsupported models."""

    meta = CostMetadata(
        input_tokens=2000,
        output_tokens=1000,
        cached_tokens=50,
        context_cache_tokens=5000,
        context_cache_hours=1,
        google=GoogleMetadata(use_vertex_ai=True, grounding_requests=1500),
        batch_mode=True,
    )
    model = "unsupported"
    assert calculate_cost(meta, model) is None



"""This module contains the `GroqCallResponse` class.

usage docs: learn/calls.md#handling-responses
"""

from collections.abc import Sequence
from functools import cached_property
from typing import cast

from groq.types.chat import (
    ChatCompletion,
    ChatCompletionAssistantMessageParam,
    ChatCompletionMessageParam,
    ChatCompletionToolMessageParam,
    ChatCompletionToolParam,
    ChatCompletionUserMessageParam,
)
from groq.types.completion_usage import CompletionUsage
from pydantic import SerializeAsAny, computed_field

from .. import BaseMessageParam
from ..base import BaseCallResponse, transform_tool_outputs
from ..base.types import CostMetadata, FinishReason, ImageMetadata
from ._utils._message_param_converter import GroqMessageParamConverter
from .call_params import GroqCallParams
from .dynamic_config import AsyncGroqDynamicConfig, GroqDynamicConfig
from .tool import GroqTool


class GroqCallResponse(
    BaseCallResponse[
        ChatCompletion,
        GroqTool,
        ChatCompletionToolParam,
        AsyncGroqDynamicConfig | GroqDynamicConfig,
        ChatCompletionMessageParam,
        GroqCallParams,
        ChatCompletionUserMessageParam,
        GroqMessageParamConverter,
    ]
):
    """A convenience wrapper around the Groq `ChatCompletion` response.

    When calling the Groq API using a function decorated with `groq_call`, the
    response will be an `GroqCallResponse` instance with properties that allow for
    more convenience access to commonly used attributes.

    Example:

    ```python
    from mirascope.core import prompt_template
    from mirascope.core.groq import groq_call


    @groq_call("llama-3.1-8b-instant")
    def recommend_book(genre: str) -> str:
        return f"Recommend a {genre} book"

    response = recommend_book("fantasy")  # response is an `GroqCallResponse` instance
    print(response.content)
    ```
    """

    _message_converter: type[GroqMessageParamConverter] = GroqMessageParamConverter

    _provider = "groq"

    @computed_field
    @property
    def content(self) -> str:
        """Returns the content of the chat completion for the 0th choice."""
        message = self.response.choices[0].message
        return message.content if message.content is not None else ""

    @computed_field
    @property
    def finish_reasons(self) -> list[str]:
        """Returns the finish reasons of the response."""
        return [str(choice.finish_reason) for choice in self.response.choices]

    @computed_field
    @property
    def model(self) -> str:
        """Returns the name of the response model."""
        return self.response.model

    @computed_field
    @property
    def id(self) -> str:
        """Returns the id of the response."""
        return self.response.id

    @property
    def usage(self) -> CompletionUsage | None:
        """Returns the usage of the chat completion."""
        return self.response.usage

    @computed_field
    @property
    def input_tokens(self) -> int | None:
        """Returns the number of input tokens."""
        return self.usage.prompt_tokens if self.usage else None

    @computed_field
    @property
    def cached_tokens(self) -> int | None:
        """Returns the number of cached tokens."""
        return 0

    @computed_field
    @property
    def output_tokens(self) -> int | None:
        """Returns the number of output tokens."""
        return self.usage.completion_tokens if self.usage else None

    @computed_field
    @cached_property
    def message_param(self) -> SerializeAsAny[ChatCompletionAssistantMessageParam]:
        """Returns the assistants's response as a message parameter."""
        message_param = self.response.choices[0].message.model_dump(
            exclude={"function_call"}
        )
        return ChatCompletionAssistantMessageParam(**message_param)

    @cached_property
    def tools(self) -> list[GroqTool] | None:
        """Returns any available tool calls as their `GroqTool` definition.

        Raises:
            ValidationError: if a tool call doesn't match the tool's schema.
        """
        tool_calls = self.response.choices[0].message.tool_calls
        if not self.tool_types or not tool_calls:
            return None

        extracted_tools = []
        for tool_call in tool_calls:
            for tool_type in self.tool_types:
                if tool_call.function.name == tool_type._name():
                    extracted_tools.append(tool_type.from_tool_call(tool_call))
                    break

        return extracted_tools

    @cached_property
    def tool(self) -> GroqTool | None:
        """Returns the 0th tool for the 0th choice message.

        Raises:
            ValidationError: if the tool call doesn't match the tool's schema.
        """
        if tools := self.tools:
            return tools[0]
        return None

    @classmethod
    @transform_tool_outputs
    def tool_message_params(
        cls, tools_and_outputs: Sequence[tuple[GroqTool, str]]
    ) -> list[ChatCompletionToolMessageParam]:
        """Returns the tool message parameters for tool call results.

        Args:
            tools_and_outputs: The sequence of tools and their outputs from which the tool
                message parameters should be constructed.

        Returns:
            The list of constructed `ChatCompletionToolMessageParam` parameters.
        """
        return [
            ChatCompletionToolMessageParam(  # pyright: ignore [reportCallIssue]
                role="tool",
                content=output,
                tool_call_id=tool.tool_call.id,  # pyright: ignore [reportOptionalMemberAccess]
                name=tool._name(),  # pyright: ignore [reportCallIssue]
            )
            for tool, output in tools_and_outputs
        ]

    @property
    def common_finish_reasons(self) -> list[FinishReason] | None:
        return cast(list[FinishReason], self.finish_reasons)

    @property
    def common_message_param(self) -> BaseMessageParam:
        return GroqMessageParamConverter.from_provider([self.message_param])[0]

    @property
    def common_user_message_param(self) -> BaseMessageParam | None:
        if not self.user_message_param:
            return None
        return GroqMessageParamConverter.from_provider([self.user_message_param])[0]

    @property
    def cost_metadata(self) -> CostMetadata:
        cost_metadata = super().cost_metadata
        image_metadata = []

        for message in self.messages:
            if "content" not in message:
                continue
            content = message["content"]

            if not isinstance(content, list):
                continue
            for part in content:
                # Check if this part is an image_url
                if isinstance(part, dict) and part.get("type") == "image_url":
                    # Only count the image if it has a URL
                    image_metadata.append(ImageMetadata(width=0, height=0))

        cost_metadata.images = image_metadata
        return cost_metadata



"""This module contains the setup_call function for Azure tools."""

import inspect
import os
import warnings
from collections.abc import AsyncIterable, Awaitable, Callable, Iterable
from typing import Any, cast, overload

from azure.ai.inference import ChatCompletionsClient
from azure.ai.inference.aio import ChatCompletionsClient as AsyncChatCompletionsClient
from azure.ai.inference.models import (
    ChatCompletions,
    ChatRequestMessage,
    StreamingChatCompletionsUpdate,
    UserMessage,
)
from azure.core.credentials import AzureKeyCredential
from pydantic import BaseModel

from ...base import BaseMessageParam, BaseTool, _utils
from ...base._utils import (
    DEFAULT_TOOL_DOCSTRING,
    AsyncCreateFn,
    CreateFn,
    get_async_create_fn,
    get_create_fn,
)
from ...base.call_params import CommonCallParams
from ...base.stream_config import StreamConfig
from .._call_kwargs import AzureCallKwargs
from ..call_params import AzureCallParams, ResponseFormatJSON
from ..dynamic_config import AsyncAzureDynamicConfig, AzureDynamicConfig
from ..tool import AzureTool, GenerateAzureStrictToolJsonSchema
from ._convert_common_call_params import convert_common_call_params
from ._convert_message_params import convert_message_params
from ._get_credential import get_credential

try:
    from azure.ai.inference.models._models import (
        ChatCompletionsResponseFormatJSON as FormatJSON,
    )

    json_object: ResponseFormatJSON | None = None
except ImportError:  # pragma: no cover
    # https://github.com/Azure/azure-sdk-for-python/commit/cf61ada8243d619ee6164bf1d32b3939b3c271e5#diff-0fb29d8cce6df3a014592ba8e9959409b8c471088381549807239324f5e8ddc4R15-R17
    from azure.ai.inference.models._models import (
        JsonSchemaFormat as FormatJSON,  # pyright: ignore [reportAttributeAccessIssue]
    )

    json_object = "json_object"  # pyright: ignore [reportAssignmentType]


@overload
def setup_call(
    *,
    model: str,
    client: AsyncChatCompletionsClient | None,
    fn: Callable[..., Awaitable[AsyncAzureDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: AsyncAzureDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: AzureCallParams | CommonCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    AsyncCreateFn[ChatCompletions, StreamingChatCompletionsUpdate],
    str | None,
    list[ChatRequestMessage],
    list[type[AzureTool]] | None,
    AzureCallKwargs,
]: ...


@overload
def setup_call(
    *,
    model: str,
    client: ChatCompletionsClient | None,
    fn: Callable[..., AzureDynamicConfig],
    fn_args: dict[str, Any],
    dynamic_config: AzureDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: AzureCallParams | CommonCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    CreateFn[ChatCompletions, StreamingChatCompletionsUpdate],
    str | None,
    list[ChatRequestMessage],
    list[type[AzureTool]] | None,
    AzureCallKwargs,
]: ...


def setup_call(
    *,
    model: str,
    client: ChatCompletionsClient | AsyncChatCompletionsClient | None,
    fn: Callable[..., AzureDynamicConfig]
    | Callable[..., Awaitable[AzureDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: AzureDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: AzureCallParams | CommonCallParams,
    response_model: type[BaseModel] | None,
    stream: bool | StreamConfig,
) -> tuple[
    CreateFn[ChatCompletions, StreamingChatCompletionsUpdate]
    | AsyncCreateFn[ChatCompletions, StreamingChatCompletionsUpdate],
    str | None,
    list[ChatRequestMessage],
    list[type[AzureTool]] | None,
    AzureCallKwargs,
]:
    prompt_template, messages, tool_types, base_call_kwargs = _utils.setup_call(
        fn,
        fn_args,
        dynamic_config,
        tools,
        AzureTool,
        call_params,
        convert_common_call_params,
    )
    call_kwargs = cast(AzureCallKwargs, base_call_kwargs)
    messages = cast(list[BaseMessageParam | ChatRequestMessage], messages)
    messages = convert_message_params(messages)
    if json_mode:
        if response_model and response_model.model_config.get("strict", False):
            call_kwargs["response_format"] = FormatJSON(
                {
                    "name": response_model.__name__,
                    "description": response_model.__doc__ or DEFAULT_TOOL_DOCSTRING,
                    "strict": True,
                    "schema": response_model.model_json_schema(
                        schema_generator=GenerateAzureStrictToolJsonSchema
                    ),
                }
            )
        else:
            call_kwargs["response_format"] = json_object or FormatJSON()
            json_mode_content = _utils.json_mode_content(response_model).strip()
            messages.append(UserMessage(content=json_mode_content))
    elif response_model:
        assert tool_types, "At least one tool must be provided for extraction."
        if tool_types and tool_types[0].model_config.get("strict", False):
            warnings.warn(
                "You must set `json_mode=True` to use `strict=True` response models. "
                "Ignoring `strict` and using tools for extraction.",
                UserWarning,
            )
        call_kwargs["tool_choice"] = "required"
    call_kwargs |= {"model": model, "messages": messages}

    if client is None:
        endpoint = os.environ["AZURE_INFERENCE_ENDPOINT"]
        credential = cast(AzureKeyCredential, get_credential())

        client = (
            AsyncChatCompletionsClient(endpoint=endpoint, credential=credential)
            if inspect.iscoroutinefunction(fn)
            else ChatCompletionsClient(endpoint=endpoint, credential=credential)
        )
    create = (
        get_async_create_fn(
            cast(
                AsyncCreateFn[
                    ChatCompletions, AsyncIterable[StreamingChatCompletionsUpdate]
                ],
                client.complete,
            )
        )
        if isinstance(client, AsyncChatCompletionsClient)
        else get_create_fn(
            cast(
                CreateFn[ChatCompletions, Iterable[StreamingChatCompletionsUpdate]],
                client.complete,
            )
        )
    )
    return create, prompt_template, messages, tool_types, call_kwargs



"""Tests the `gemini.call_response_chunk` module."""

from google.ai.generativelanguage import (
    Candidate,
    Content,
    GenerateContentResponse,
    Part,
)
from google.generativeai.types import (  # type: ignore
    GenerateContentResponse as GenerateContentResponseType,
)

from mirascope.core.base.types import CostMetadata
from mirascope.core.gemini.call_response_chunk import GeminiCallResponseChunk


def test_gemini_call_response_chunk() -> None:
    """Tests the `GeminiCallResponseChunk` class."""
    chunk = GenerateContentResponseType.from_response(
        GenerateContentResponse(
            candidates=[
                Candidate(
                    finish_reason=1,
                    content=Content(
                        parts=[Part(text="The author is Patrick Rothfuss")],
                        role="model",
                    ),
                )
            ]
        )
    )
    call_response_chunk = GeminiCallResponseChunk(chunk=chunk)
    assert call_response_chunk.content == "The author is Patrick Rothfuss"
    assert call_response_chunk.finish_reasons == [Candidate.FinishReason.STOP]
    assert call_response_chunk.model is None
    assert call_response_chunk.id is None
    assert call_response_chunk.usage is None
    assert call_response_chunk.input_tokens is None
    assert call_response_chunk.output_tokens is None
    assert call_response_chunk.common_finish_reasons == ["stop"]
    assert call_response_chunk.cost_metadata == CostMetadata()



"""Tests the `mistral._utils.get_json_output` module."""

import pytest
from mistralai.models import (
    AssistantMessage,
    ChatCompletionChoice,
    ChatCompletionResponse,
    CompletionChunk,
    CompletionResponseStreamChoice,
    DeltaMessage,
    FunctionCall,
    ToolCall,
    UsageInfo,
)

from mirascope.core.mistral._utils._get_json_output import get_json_output
from mirascope.core.mistral.call_response import MistralCallResponse
from mirascope.core.mistral.call_response_chunk import MistralCallResponseChunk


def test_get_json_output_call_response() -> None:
    """Tests the `get_json_output` function with a call response."""

    tool_call = ToolCall(
        id="id",
        function=FunctionCall(
            name="FormatBook",
            arguments='{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
        ),
        type="function",
    )
    choices = [
        ChatCompletionChoice(
            index=0,
            message=AssistantMessage(content="json_output", tool_calls=[tool_call]),
            finish_reason="stop",
        )
    ]
    completion = ChatCompletionResponse(
        id="id",
        object="",
        created=0,
        model="mistral-large-latest",
        choices=choices,
        usage=UsageInfo(prompt_tokens=1, completion_tokens=1, total_tokens=2),
    )
    call_response = MistralCallResponse(
        metadata={},
        response=completion,
        tool_types=None,
        prompt_template="",
        fn_args={},
        dynamic_config=None,
        messages=[],
        call_params={},
        call_kwargs={},
        user_message_param=None,
        start_time=0,
        end_time=0,
    )
    assert get_json_output(call_response, json_mode=True) == "json_output"
    assert (
        get_json_output(call_response, json_mode=False)
        == '{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}'
    )

    completion.choices[0].message.content = ""  # pyright: ignore [reportOptionalSubscript]
    completion.choices[0].message.tool_calls = None  # pyright: ignore [reportOptionalSubscript]
    with pytest.raises(
        ValueError, match="No tool call or JSON object found in response."
    ):
        get_json_output(call_response, json_mode=False)


def test_get_json_output_call_response_chunk() -> None:
    """Tests the `get_json_output` function with a call response chunk."""
    tool_call = ToolCall(
        id="id",
        function=FunctionCall(
            arguments='{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}',
            name="function",
        ),
        type="function",
    )
    choices = [
        CompletionResponseStreamChoice(
            index=0,
            delta=DeltaMessage(content="json_output", tool_calls=[tool_call]),
            finish_reason=None,
        )
    ]
    chunk = CompletionChunk(
        id="id",
        model="mistral-large-latest",
        choices=choices,
        created=0,
        object=None,
        usage=None,
    )
    call_response_chunk = MistralCallResponseChunk(chunk=chunk)
    assert get_json_output(call_response_chunk, json_mode=True) == "json_output"
    assert (
        get_json_output(call_response_chunk, json_mode=False)
        == '{"title": "The Name of the Wind", "author": "Patrick Rothfuss"}'
    )

    chunk.choices[0].delta.tool_calls = None
    assert get_json_output(call_response_chunk, json_mode=False) == ""



from collections.abc import Callable
from functools import reduce, wraps
from typing import Any, ParamSpec, TypeVar

_P = ParamSpec("_P")
_R = TypeVar("_R")

_WP = ParamSpec("_WP")
_WR = TypeVar("_WR")


def merge_decorators(
    decorator: Callable[[Callable[_P, _R]], Callable[_WP, _WR]],
    *additional_decorators: Callable[[Callable], Callable],
) -> Callable[[], Callable[[Callable[_P, _R]], Callable[_WP, _WR]]]:
    """Combines multiple decorators into a single decorator factory.

    This function allows you to merge multiple decorators into a single decorator factory.
    The decorators are applied in the order they are passed to the function.
    All function metadata (e.g. docstrings, function name) is preserved through the decoration chain.

    Args:
        decorator: The base decorator that determines the type signature of the decorated function.
        *additional_decorators: Additional decorators to be merged with the base decorator.

    Returns:
        A decorator factory function that applies all decorators in the specified order.

    Example:
        ```python
        # Combine multiple decorators
        merged = merge_decorators(
            retry(),
            with_logging(),
            handle_errors()
        )

        # Apply the merged decorators
        @merged()
        def my_function():
            pass
        ```
    """
    decorators = [decorator] + list(additional_decorators)

    def decorator_factory() -> Callable[[Callable[_P, _R]], Callable[_WP, _WR]]:
        def inner(func: Callable[_P, _R]) -> Callable[_WP, _WR]:
            def compose(f: Callable, d: Callable) -> Callable:
                @wraps(f)
                def wrapped(*args: Any, **kwargs: Any) -> Any:  # noqa: ANN401
                    return d(f)(*args, **kwargs)

                return wrapped

            return reduce(compose, reversed(decorators), func)  # pyright: ignore [reportReturnType]

        return inner

    return decorator_factory



"""Tests for the `base_prompt` module."""

import os
from typing import ClassVar
from unittest import mock
from unittest.mock import AsyncMock, MagicMock

import pytest
from pydantic import computed_field

from mirascope.core import BaseMessageParam
from mirascope.core.base.prompt import BasePrompt, metadata, prompt_template


def test_base_prompt() -> None:
    """Tests the `BasePrompt` class."""

    @prompt_template("Recommend a {genre} book.")
    class BookRecommendationPrompt(BasePrompt):
        genre: str

    prompt = BookRecommendationPrompt(genre="fantasy")
    assert str(prompt) == "Recommend a fantasy book."
    assert prompt.dump() == {
        "metadata": {},
        "prompt": "Recommend a fantasy book.",
        "template": "Recommend a {genre} book.",
        "inputs": {"genre": "fantasy"},
    }

    @prompt_template(
        """
        SYSTEM: You are a helpful assistant.
        USER: Please help me.
        """
    )
    class MessagesPrompt(BasePrompt): ...

    prompt = MessagesPrompt()
    assert prompt.message_params() == [
        BaseMessageParam(role="system", content="You are a helpful assistant."),
        BaseMessageParam(role="user", content="Please help me."),
    ]


def test_base_prompt_with_computed_fields() -> None:
    """Tests the `BasePrompt` class with list and list[list] computed fields."""

    @prompt_template("Recommend a {genre} book.")
    class BookRecommendationPrompt(BasePrompt):
        @computed_field
        @property
        def genre(self) -> str:
            return "fantasy"

    prompt = BookRecommendationPrompt()
    assert str(prompt) == "Recommend a fantasy book."


def test_base_prompt_with_prompt_template() -> None:
    """Tests the `BasePrompt` class with prompt_template."""

    class BookRecommendationPrompt(BasePrompt):
        prompt_template: ClassVar[str] = "Recommend a {genre} book."
        genre: str

    prompt = BookRecommendationPrompt(genre="fantasy")
    assert str(prompt) == "Recommend a fantasy book."
    assert prompt.dump() == {
        "metadata": {},
        "prompt": "Recommend a fantasy book.",
        "template": "Recommend a {genre} book.",
        "inputs": {"genre": "fantasy"},
    }

    class BookRecommendationPromptWithoutClassVar(BasePrompt):
        prompt_template = "Recommend a {genre} book."
        genre: str

    prompt = BookRecommendationPromptWithoutClassVar(genre="fantasy")
    assert str(prompt) == "Recommend a fantasy book."
    assert prompt.dump() == {
        "metadata": {},
        "prompt": "Recommend a fantasy book.",
        "template": "Recommend a {genre} book.",
        "inputs": {"genre": "fantasy"},
    }

    class MessagesPrompt(BasePrompt):
        prompt_template: ClassVar[str] = """
        SYSTEM: You are a helpful assistant.
        USER: Please help me.
        """

    prompt = MessagesPrompt()
    assert prompt.message_params() == [
        BaseMessageParam(role="system", content="You are a helpful assistant."),
        BaseMessageParam(role="user", content="Please help me."),
    ]

    class MessagesPromptWithoutClassVar(BasePrompt):
        prompt_template = """
        SYSTEM: You are a helpful assistant.
        USER: Please help me.
        """

    prompt = MessagesPromptWithoutClassVar()
    assert prompt.message_params() == [
        BaseMessageParam(role="system", content="You are a helpful assistant."),
        BaseMessageParam(role="user", content="Please help me."),
    ]


def test_base_prompt_run() -> None:
    """Tests the `BasePrompt.run` method."""
    mock_decorator = MagicMock()
    mock_call_fn = MagicMock()
    mock_decorator.return_value = mock_call_fn
    mock_call_fn.return_value = "response"

    @prompt_template("Recommend a {genre} book.")
    class BookRecommendationPrompt(BasePrompt):
        genre: str

    prompt = BookRecommendationPrompt(genre="fantasy")
    response = prompt.run(mock_decorator)
    assert response == mock_call_fn.return_value

    # Ensure the decorator was called with the correct function
    mock_decorator.assert_called_once()
    decorator_arg = mock_decorator.call_args[0][0]
    assert callable(decorator_arg)
    assert hasattr(decorator_arg, "_prompt_template")
    assert decorator_arg._prompt_template == "Recommend a {genre} book."  # pyright: ignore [reportFunctionMemberAccess]

    # Ensure the decorated function was called with the correct arguments
    mock_call_fn.assert_called_once_with(genre="fantasy")


@pytest.mark.asyncio
async def test_base_prompt_run_async() -> None:
    mock_decorator = MagicMock()
    mock_call_fn = AsyncMock()
    mock_decorator.return_value = mock_call_fn
    mock_call_fn.return_value = "response"

    @prompt_template("Recommend a {genre} book.")
    class BookRecommendationPrompt(BasePrompt):
        genre: str

    prompt = BookRecommendationPrompt(genre="fantasy")
    response = await prompt.run_async(mock_decorator)
    assert response == mock_call_fn.return_value

    # Ensure the decorator was called with the correct function
    mock_decorator.assert_called_once()
    decorator_arg = mock_decorator.call_args[0][0]
    assert callable(decorator_arg)
    assert hasattr(decorator_arg, "_prompt_template")
    assert decorator_arg._prompt_template == "Recommend a {genre} book."  # pyright: ignore [reportFunctionMemberAccess]

    # Ensure the decorated function was called with the correct arguments
    mock_call_fn.assert_called_once_with(genre="fantasy")


def test_prompt_template_docstring() -> None:
    """Tests the `prompt_template` decorator on a `BasePrompt`."""

    os.environ["MIRASCOPE_DOCSTRING_PROMPT_TEMPLATE"] = "ENABLED"

    class BookRecommendationPrompt(BasePrompt):
        """Recommend a {genre} book."""

        genre: str

    prompt = BookRecommendationPrompt(genre="fantasy")
    assert str(prompt) == "Recommend a fantasy book."

    os.environ["MIRASCOPE_DOCSTRING_PROMPT_TEMPLATE"] = "DISABLED"


def test_prompt_template_with_function() -> None:
    """Tests the `prompt_template` decorator on a function."""

    @prompt_template("Recommend a {genre} book.")
    def fn(genre: str) -> None: ...

    assert (
        hasattr(fn, "_prompt_template")
        and fn._prompt_template == "Recommend a {genre} book."  # pyright: ignore [reportFunctionMemberAccess]
    )
    assert fn("fantasy") == [
        BaseMessageParam(role="user", content="Recommend a fantasy book.")
    ]
    assert fn(genre="fantasy") == [
        BaseMessageParam(role="user", content="Recommend a fantasy book.")
    ]


@pytest.mark.asyncio
async def test_prompt_template_with_async_function() -> None:
    """Tests the `prompt_template` decorator on a async function."""

    @prompt_template("Recommend a {genre} book.")
    async def fn(genre: str) -> None: ...

    assert (
        hasattr(fn, "_prompt_template")
        and fn._prompt_template == "Recommend a {genre} book."  # pyright: ignore [reportFunctionMemberAccess]
    )
    assert await fn("fantasy") == [
        BaseMessageParam(role="user", content="Recommend a fantasy book.")
    ]
    assert await fn(genre="fantasy") == [
        BaseMessageParam(role="user", content="Recommend a fantasy book.")
    ]


def test_metadata_decorator() -> None:
    """Tests the `metadata` decorator on a `BasePrompt`."""

    @metadata({"tags": {"version:0001"}})
    @prompt_template("Recommend a book.")
    class BookRecommendationPrompt(BasePrompt): ...

    prompt = BookRecommendationPrompt()
    assert prompt.dump()["metadata"] == {"tags": {"version:0001"}}

    @metadata({"tags": {"version:0001"}})
    def fn() -> None: ...

    assert hasattr(fn, "_metadata") and fn._metadata == {"tags": {"version:0001"}}  # pyright: ignore [reportFunctionMemberAccess]


def test_prompt_template_with_none() -> None:
    """Tests the `prompt_template` decorator with `None` arguments."""
    with mock.patch("mirascope.core.base.prompt.messages_decorator") as mock_decorator:
        mock_decorated_function = mock.MagicMock()
        mock_decorator.return_value.return_value = mock_decorated_function

        @prompt_template()
        def fn() -> None: ...

        assert fn == mock_decorated_function


def test_base_prompt_str_with_special_tags() -> None:
    """Tests the BasePrompt.__str__ method with special tags."""

    @prompt_template(
        """
        Process this: {content:text} and {image:image} and {audio:audio} and 
        these: {contents:texts} and {images:images} and {audios:audios}
        """
    )
    class MultiModalPrompt(BasePrompt):
        content: str
        image: bytes
        audio: bytes
        contents: list[str]
        images: list[bytes]
        audios: list[bytes]

    prompt = MultiModalPrompt(
        content="text",
        image=b"image",
        audio=b"audio",
        contents=["text1", "text2"],
        images=[b"image1", b"image2"],
        audios=[b"audio1", b"audio2"],
    )

    # Verify all special tags are removed in string representation
    expected = (
        "Process this: text and b'image' and b'audio' and \n"
        "these: ['text1', 'text2'] and [b'image1', b'image2'] and [b'audio1', b'audio2']"
    )
    assert str(prompt).strip() == expected.strip()



import json

from groq.types.chat import (
    ChatCompletionAssistantMessageParam,
    ChatCompletionToolMessageParam,
    ChatCompletionUserMessageParam,
)

from mirascope.core import BaseMessageParam
from mirascope.core.base import ImageURLPart, TextPart, ToolCallPart, ToolResultPart
from mirascope.core.groq._utils._message_param_converter import (
    GroqMessageParamConverter,
)


def test_convert_with_only_content():
    """
    If there's only content string, role="assistant", content=that string.
    """
    message_param = ChatCompletionAssistantMessageParam(
        role="assistant", content="Hello world"
    )
    results = GroqMessageParamConverter.from_provider([message_param])
    assert len(results) == 1

    result = results[0]
    assert isinstance(result, BaseMessageParam)
    assert result.role == "assistant"
    assert result.content == "Hello world"


def test_convert_content_list():
    """Test that we can convert a list of content parts properly."""
    message_param = ChatCompletionUserMessageParam(
        role="user",
        content=[
            {"type": "text", "text": "content"},
            {
                "type": "image_url",
                "image_url": {"url": "https://example.com/image.jpg"},
            },
        ],
    )
    results = GroqMessageParamConverter.from_provider([message_param])
    assert len(results) == 1

    result = results[0]
    assert result.role == "user"
    assert result.content == [
        TextPart(type="text", text="content"),
        ImageURLPart(
            type="image_url", url="https://example.com/image.jpg", detail=None
        ),
    ]


def test_convert_with_content_and_tool_calls():
    """
    If there's both content and tool_calls, final role="assistant" and we
    have [ToolCallPart].
    """
    message_param = ChatCompletionAssistantMessageParam(
        role="assistant",
        content="This is a message",
        tool_calls=[
            {
                "type": "function",
                "function": {
                    "name": "test_tool",
                    "arguments": json.dumps({"param": "value"}),
                },
                "id": "tool_call_123",
            }
        ],
    )
    results = GroqMessageParamConverter.from_provider([message_param])
    assert len(results) == 1

    result = results[0]
    assert isinstance(result, BaseMessageParam)
    assert result.role == "assistant"
    assert len(result.content) == 1
    assert isinstance(result.content[0], ToolCallPart)
    assert result.content[0].name == "test_tool"
    assert result.content[0].args == {"param": "value"}
    assert result.content[0].id == "tool_call_123"


def test_convert_with_only_tool_calls():
    """
    If there's no content but tool_calls, we only get ToolCallParts in the final list.
    """
    message_param = ChatCompletionAssistantMessageParam(
        role="assistant",
        content=None,
        tool_calls=[
            {
                "type": "function",
                "function": {
                    "name": "some_tool",
                    "arguments": json.dumps({"key": "val"}),
                },
                "id": "tool_call_456",
            }
        ],
    )
    results = GroqMessageParamConverter.from_provider([message_param])
    assert len(results) == 1

    result = results[0]
    assert result.role == "assistant"
    assert len(result.content) == 1
    part = result.content[0]
    assert isinstance(part, ToolCallPart)
    assert part.name == "some_tool"
    assert part.args == {"key": "val"}
    assert part.id == "tool_call_456"


def test_convert_with_no_content_no_tool_calls():
    """
    If there's no content or tool_calls, we might get role="tool" with empty content or role="assistant".
    Depends on your internal logic.
    """
    message_param = ChatCompletionAssistantMessageParam(
        role="assistant", content=None, tool_calls=[]
    )
    results = GroqMessageParamConverter.from_provider([message_param])
    assert len(results) == 1

    result = results[0]
    assert result.role == "assistant"
    assert result.content == ""


def test_convert_with_tool_result():
    """Tests that we properly convert tool messages."""
    message_param = ChatCompletionToolMessageParam(
        role="tool", tool_call_id="id", content="output"
    )
    results = GroqMessageParamConverter.from_provider([message_param])
    assert results == [
        BaseMessageParam(
            role="tool",
            content=[
                ToolResultPart(
                    type="tool_result",
                    name="",
                    content="output",
                    id="id",
                    is_error=False,
                )
            ],
        )
    ]


def test_convert_base_message_param():
    """If there's a BaseMessageParam, we should get that as is."""
    message_param = BaseMessageParam(role="assistant", content="Hello world")
    results = GroqMessageParamConverter.to_provider([message_param])
    assert results == [
        ChatCompletionAssistantMessageParam(role="assistant", content="Hello world")
    ]



"""Utility for converting `BaseMessageParam` to `ContentsType`"""

import io

import PIL.Image
from google.generativeai import protos
from google.generativeai.types import ContentDict

from ...base import BaseMessageParam
from ...base._utils import get_audio_type
from ...base._utils._parse_content_template import _load_media


def convert_message_params(
    message_params: list[BaseMessageParam | ContentDict],
) -> list[ContentDict]:
    converted_message_params = []
    for message_param in message_params:
        if not isinstance(message_param, BaseMessageParam):
            converted_message_params.append(message_param)
        elif (role := message_param.role) == "system":
            if not isinstance(message_param.content, str):
                raise ValueError(
                    "System message content must be a single text string."
                )  # pragma: no cover
            converted_message_params += [
                {
                    "role": "system",
                    "parts": [message_param.content],
                },
            ]
        elif isinstance((content := message_param.content), str):
            converted_message_params.append(
                {"role": role if role == "user" else "model", "parts": [content]}
            )
        else:
            converted_content = []
            for part in content:
                if part.type == "text":
                    converted_content.append(part.text)
                elif part.type == "image":
                    if part.media_type not in [
                        "image/jpeg",
                        "image/png",
                        "image/webp",
                        "image/heic",
                        "image/heif",
                    ]:
                        raise ValueError(
                            f"Unsupported image media type: {part.media_type}. "
                            "Gemini currently only supports JPEG, PNG, WebP, HEIC, "
                            "and HEIF images."
                        )
                    image = PIL.Image.open(io.BytesIO(part.image))
                    converted_content.append(image)
                elif part.type == "image_url":
                    if part.url.startswith(("https://", "http://")):
                        image = PIL.Image.open(io.BytesIO(_load_media(part.url)))
                        media_type = (
                            PIL.Image.MIME[image.format]
                            if image.format
                            else "image/unknown"
                        )
                        if media_type not in [
                            "image/jpeg",
                            "image/png",
                            "image/webp",
                            "image/heic",
                            "image/heif",
                        ]:
                            raise ValueError(
                                f"Unsupported image media type: {media_type}. "
                                "Gemini currently only supports JPEG, PNG, WebP, HEIC, "
                                "and HEIF images."
                            )
                        converted_content.append(image)
                    else:
                        converted_content.append(protos.FileData(file_uri=part.url))
                elif part.type == "audio":
                    if part.media_type not in [
                        "audio/wav",
                        "audio/mp3",
                        "audio/aiff",
                        "audio/aac",
                        "audio/ogg",
                        "audio/flac",
                    ]:
                        raise ValueError(
                            f"Unsupported audio media type: {part.media_type}. "
                            "Gemini currently only supports WAV, MP3, AIFF, AAC, OGG, "
                            "and FLAC audio file types."
                        )
                    converted_content.append(
                        {"mime_type": part.media_type, "data": part.audio}
                    )
                elif part.type == "audio_url":
                    if part.url.startswith(("https://", "http://")):
                        audio = _load_media(part.url)
                        audio_type = f"audio/{get_audio_type(audio)}"
                        if audio_type not in [
                            "audio/wav",
                            "audio/mp3",
                            "audio/aiff",
                            "audio/aac",
                            "audio/ogg",
                            "audio/flac",
                        ]:
                            raise ValueError(
                                f"Unsupported audio media type: {audio_type}. "
                                "Gemini currently only supports WAV, MP3, AIFF, AAC, OGG, "
                                "and FLAC audio file types."
                            )
                        converted_content.append(
                            {"mime_type": audio_type, "data": audio}
                        )
                    else:
                        converted_content.append(protos.FileData(file_uri=part.url))
                elif part.type == "tool_call":
                    converted_content.append(
                        protos.FunctionCall(
                            name=part.name,
                            args=part.args,
                        )
                    )
                elif part.type == "tool_result":
                    if converted_content:
                        converted_message_params.append(
                            {
                                "role": role if role == "user" else "model",
                                "parts": converted_content,
                            }
                        )
                        converted_content = []
                    converted_message_params.append(
                        {
                            "role": "user",
                            "parts": [
                                protos.FunctionResponse(
                                    name=part.name, response={"result": part.content}
                                )
                            ],
                        }
                    )
                else:
                    raise ValueError(
                        "Gemini currently only supports text, image, and audio parts. "
                        f"Part provided: {part.type}"
                    )
            if converted_content:
                converted_message_params.append(
                    {
                        "role": role if role == "user" else "model",
                        "parts": converted_content,
                    }
                )
    return converted_message_params



from typing import ClassVar, Literal

import httpx
from pydantic import Field

from mirascope.tools.base import ConfigurableTool, _ConfigurableToolConfig


class HTTPXConfig(_ConfigurableToolConfig):
    """Configuration for HTTPX requests"""

    timeout: int = Field(
        default=5,
        description="Request timeout in seconds. When None, no timeout will be applied",
    )


class _BaseHTTPX(ConfigurableTool):
    """Tool for making HTTP requests using HTTPX with configurable timeout and error handling."""

    __configurable_tool_config__ = HTTPXConfig()

    __prompt_usage_description__: ClassVar[str] = """
    - `HTTPX`: Makes HTTP requests to web URLs with HTTPX client
        - Supports standard HTTP methods (GET, POST, PUT, DELETE, etc.)
        - Allows configuration of request data, headers, and parameters
        - Automatically handles encoding/decoding of request/response content
        - Handles redirects automatically (configurable)
        - Returns response text or error message on failure
        - Configurable timeout for requests
    """

    url: str = Field(..., description="URL to request")
    method: Literal["GET", "POST", "PUT", "DELETE", "HEAD", "OPTIONS"] = Field(
        "GET", description="HTTP method to use"
    )
    data: dict | None = Field(
        None, description="Form data to send with POST/PUT requests"
    )
    json_: dict | None = Field(
        None, description="JSON data to send with POST/PUT requests", alias="json"
    )
    params: dict | None = Field(
        None, description="URL parameters to include in the request"
    )
    headers: dict | None = Field(None, description="Request headers")
    follow_redirects: bool = Field(
        True, description="Whether to follow redirects automatically"
    )


class HTTPX(_BaseHTTPX):
    def call(self) -> str:
        """
        Make an HTTP request to the given URL using HTTPX.

        Returns:
            str: Response text if successful, error message if request fails
        """
        try:
            # Configure timeout - None means no timeout
            timeout = (
                httpx.Timeout(self._get_config().timeout)
                if self._get_config().timeout is not None
                else None
            )

            # Make request using context manager for proper resource cleanup
            with httpx.Client(timeout=timeout) as client:
                response = client.request(
                    method=self.method,
                    url=self.url,
                    params=self.params,
                    json=self.json_,
                    data=self.data,
                    headers=self.headers,
                    follow_redirects=self.follow_redirects,
                )
                response.raise_for_status()
                return response.text

        except httpx.RequestError as e:
            return f"Request error occurred: {str(e)}"
        except httpx.HTTPStatusError as e:
            return f"HTTP error occurred: {e.response.status_code} - {str(e)}"
        except Exception as e:
            return f"{type(e).__name__}: Failed to make request to {self.url}"


class AsyncHTTPX(_BaseHTTPX):
    async def call(self) -> str:
        """
        Make an asynchronous HTTP request to the given URL using HTTPX.

        Returns:
            str: Response text if successful, error message if request fails
        """
        try:
            # Configure timeout - None means no timeout
            timeout = (
                httpx.Timeout(self._get_config().timeout)
                if self._get_config().timeout is not None
                else None
            )

            # Make async request using async context manager
            async with httpx.AsyncClient(timeout=timeout) as client:
                response = await client.request(
                    method=self.method,
                    url=self.url,
                    params=self.params,
                    json=self.json_,
                    data=self.data,
                    headers=self.headers,
                    follow_redirects=self.follow_redirects,
                )
                response.raise_for_status()
                return response.text

        except httpx.RequestError as e:
            return f"Request error occurred: {str(e)}"
        except httpx.HTTPStatusError as e:
            return f"HTTP error occurred: {e.response.status_code} - {str(e)}"
        except Exception as e:
            return f"{type(e).__name__}: Failed to make request to {self.url}"
